"""Module containing classes for datagrid MVC implementation."""

import base64
import contextlib
import datetime
import itertools
import logging
import os

from gi.repository import (
    GLib,
    GObject,
    Gdk,
    GdkPixbuf,
    Gtk,
    Pango,
)
from pygtkcompat.generictreemodel import GenericTreeModel

from datagrid_gtk3.ui.popupcal import DateEntry
from datagrid_gtk3.ui.uifile import UIFile
from datagrid_gtk3.utils.dateutils import normalize_timestamp
from datagrid_gtk3.utils.imageutils import ImageCacheManager
from datagrid_gtk3.utils.transformations import get_transformer

_MEDIA_FILES = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    os.pardir,
    "data",
    "media"
)

logger = logging.getLogger(__name__)
_no_image_loader = GdkPixbuf.PixbufLoader.new_with_type("png")
_no_image_loader.write(base64.b64decode("""
iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAYAAAAfFcSJAAAABmJLR0QA/wD/AP+gvaeTAAAACXBI
WXMAAAsTAAALEwEAmpwYAAAAB3RJTUUH3wEPEDYaIuf2wwAAABl0RVh0Q29tbWVudABDcmVhdGVk
IHdpdGggR0lNUFeBDhcAAAANSURBVAjXY2BgYGAAAAAFAAFe8yo6AAAAAElFTkSuQmCC
"""))
_no_image_loader.close()
# A trivial 1px transparent png to be used on CellRendererPixbuf when there's
# no data there. Due to possible bug on gtk, passing None to it will make it
# repeat the lastest value read in a row for that column
NO_IMAGE_PIXBUF = _no_image_loader.get_pixbuf()

# Used to represent "no option selected" on filters. We use this instead of
# None as it can be a valid value for filtering.
NO_FILTER_OPTION = object()


class OptionsPopup(Gtk.Window):

    """Popup to select which columns should be displayed on datagrid.

    :param toggle_btn: the toggle button responsible for popping this up
    :type toggle_btn: :class:`Gtk.ToggleButton`
    :param controller: the datagrid controller
    :type controller: :class:`DataGridController`
    """

    OPTIONS_PADDING = 5
    MAX_HEIGHT = 500

    (VIEW_TREE,
     VIEW_FLAT,
     VIEW_ICON) = range(3)

    __gsignals__ = {
        'column-visibility-changed': (GObject.SignalFlags.RUN_FIRST,
                                      None, (str, bool)),
        'view-changed': (GObject.SignalFlags.RUN_FIRST, None, (int, ))
    }

    def __init__(self, toggle_btn, controller, *args, **kwargs):
        self._toggle_btn = toggle_btn
        self._toggled_id = self._toggle_btn.connect(
            'toggled', self.on_toggle_button_toggled)
        self._controller = controller

        super(OptionsPopup, self).__init__(
            Gtk.WindowType.POPUP, *args, **kwargs)

        self.connect('button-press-event', self.on_button_press_event)
        self.connect('key-press-event', self.on_key_press_event)

        self._scrolled_window = Gtk.ScrolledWindow(
            vscrollbar_policy=Gtk.PolicyType.AUTOMATIC,
            hscrollbar_policy=Gtk.PolicyType.NEVER)

        alignment = Gtk.Alignment()
        alignment.set_padding(5, 5, 5, 5)
        alignment.add(self._scrolled_window)

        self.add(alignment)

    ##
    # Public
    ##

    def popup(self):
        """Show the popup.

        This will show the popup and allow the user to change
        the columns visibility.
        """
        if not self._toggle_btn.get_realized():
            return

        child = self._scrolled_window.get_child()
        if child:
            self._scrolled_window.remove(child)

        vbox = Gtk.VBox()
        combo = self._get_view_options()
        if combo is not None:
            vbox.pack_start(combo, expand=False, fill=False,
                            padding=self.OPTIONS_PADDING)

        if not isinstance(self._controller.view, DataGridIconView):
            if combo is not None:
                vbox.pack_start(
                    Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL),
                    expand=True, fill=True, padding=self.OPTIONS_PADDING)
            for switch in self._get_visibility_options():
                vbox.pack_start(switch, expand=False, fill=False,
                                padding=self.OPTIONS_PADDING)

        self._scrolled_window.add(vbox)

        toplevel = self._toggle_btn.get_toplevel().get_toplevel()
        if isinstance(toplevel, (Gtk.Window, Gtk.Dialog)):
            group = toplevel.get_group()
            if group:
                group.add_window(self)

        x, y = self._get_position()
        self.move(x, y)
        self.show_all()

        allocation = vbox.get_allocation()
        height = min(allocation.height + 2 * self.OPTIONS_PADDING,
                     self.MAX_HEIGHT)
        self.set_size_request(-1, height)

        if not self._popup_grab_window():
            self.popdown()

    def popdown(self):
        """Hide the popup."""
        if not self._toggle_btn.get_realized():
            return

        # Make sure the toggle button is unset when popping down.
        with self._toggle_btn.handler_block(self._toggled_id):
            self._toggle_btn.set_active(False)

        self.grab_remove()
        self.hide()

    ##
    # Private
    ##

    def _popup_grab_window(self):
        """Grab pointer and keyboard on this window.

        By grabbing the pointer and the keyboard, we will be able to
        intercept key-press and button-press events.
        """
        window = self.get_window()
        grab_status = Gdk.pointer_grab(
            window, True,
            (Gdk.EventMask.BUTTON_PRESS_MASK |
             Gdk.EventMask.BUTTON_RELEASE_MASK |
             Gdk.EventMask.POINTER_MOTION_MASK),
            None, None, 0L)
        if grab_status == Gdk.GrabStatus.SUCCESS:
            if Gdk.keyboard_grab(window, True, 0L) != Gdk.GrabStatus.SUCCESS:
                display = window.get_display()
                display.pointer_ungrab(0L)
                return False

        self.grab_add()
        return True

    def _get_position(self):
        """Get the position to show this popup."""
        allocation = self._toggle_btn.get_allocation()
        window = self._toggle_btn.get_window()

        if self._toggle_btn.get_has_window():
            x_coord = 0
            y_coord = 0
        else:
            x_coord = allocation.x
            y_coord = allocation.y

        x, y = window.get_root_coords(x_coord, y_coord)

        return x, y + allocation.height

    def _get_view_options(self):
        """Build view options for datagrid."""
        iters = {}
        model = Gtk.ListStore(str, int)

        iters[self.VIEW_TREE] = model.append(("Tree View", self.VIEW_TREE))

        if self._controller.model.flat_column_idx is not None:
            iters[self.VIEW_FLAT] = model.append(("Flat View", self.VIEW_FLAT))

        if any(c['transform'] == 'image'
               for c in self._controller.model.columns):
            iters[self.VIEW_ICON] = model.append(("Icon View", self.VIEW_ICON))

        # Avoid displaying the combo if there's only one option
        if len(iters) == 1:
            return None

        combo = Gtk.ComboBox()
        combo.set_model(model)
        renderer = Gtk.CellRendererText()
        combo.pack_start(renderer, True)
        combo.add_attribute(renderer, 'text', 0)

        if isinstance(self._controller.view, DataGridView):
            if self._controller.model.active_params.get('flat', False):
                combo.set_active_iter(iters[self.VIEW_FLAT])
            else:
                combo.set_active_iter(iters[self.VIEW_TREE])
        elif isinstance(self._controller.view, DataGridIconView):
            combo.set_active_iter(iters[self.VIEW_ICON])
        else:
            raise AssertionError("Unknown view type %r" % (
                self._controller.view, ))

        combo.connect('changed', self.on_combo_view_changed)
        return combo

    def _get_visibility_options(self):
        """Construct the switches based on the actual model columns."""
        model = self._controller.model
        hidden_columns = model.hidden_columns

        for column in model.columns:
            if column['name'] in hidden_columns:
                continue

            switch = Gtk.Switch()
            label = Gtk.Label(column['display'])
            switch.set_active(column['name'] in model.display_columns)

            hbox = Gtk.HBox(spacing=5)
            hbox.pack_start(switch, expand=False, fill=True, padding=0)
            hbox.pack_start(label, expand=True, fill=True, padding=0)

            switch.connect(
                'notify::active',
                self.on_column_switch_notify_active, column['name'])

            yield hbox

    ##
    # Callbacks
    ##

    def on_key_press_event(self, window, event):
        """Handle key press events.

        Popdown when the user presses Esc.
        """
        if event.get_keyval()[1] == Gdk.KEY_Escape:
            self.popdown()
            return True
        return False

    def on_button_press_event(self, window, event):
        """Handle button press events.

        Popdown when the user clicks on an area outside this window.
        """
        event_rect = Gdk.Rectangle()
        event_rect.x, event_rect.y = event.get_root_coords()
        event_rect.width = 1
        event_rect.height = 1

        allocation = self.get_allocation()
        window_rect = Gdk.Rectangle()
        window_rect.x, window_rect.y = self._get_position()
        window_rect.width = allocation.width
        window_rect.height = allocation.height

        intersection = Gdk.rectangle_intersect(
            event_rect, window_rect)
        # if the click was outside this window, hide it
        if not intersection[0]:
            self.popdown()

    def on_combo_view_changed(self, widget):
        """Handle changes on the view combo.

        Emit 'view-changed' for the given view.

        :param widget: the combobox that received the event
        :type widget: :class:`Gtk.ComboBox`
        """
        model = widget.get_model()
        value = model[widget.get_active()][1]
        self.emit('view-changed', value)
        self.popdown()

    def on_toggle_button_toggled(self, widget):
        """Show switch list of columns to display.

        :param widget: the ToggleButton that launches the list
        :type widget: :class:`Gtk.ToggleButton`
        """
        if widget.get_active():
            self.popup()
        else:
            self.popdown()

    def on_column_switch_notify_active(self, widget, p_spec, name):
        """Set the list of columns to display based on column checkboxes.

        :param widget: checkbox widget for selected/deselected column
        :type widget: :class:`Gtk.Switch`
        :param str name: name of the column to add/remove from list
        """
        self.emit('column-visibility-changed', name, widget.get_active())


class DataGridContainer(UIFile):

    """Provides UI container for tabular data TreeStore grid.

    :param window: Window for main launching application -- needed for dialog
        interaction
    :type window: :class:`Gtk.Window`
    """

    UI_FNAME = os.path.join(
        os.path.dirname(os.path.realpath(__file__)),
        'glade',
        'datagrid.glade')

    def __init__(self, window):
        """Set up container."""
        self.window = window
        UIFile.__init__(self, self.UI_FNAME)


class DataGridController(object):

    """UI controls to manipulate datagrid model/view.

    :param container: ``UIFile`` instance providing ``Gtk.Box`` and
        access to GTK widgets for controller
    :type container: :class:`DataGridContainer`
    :param data_source: Database backend instance
    :type data_source: :class:`datagrid_gtk3.db.sqlite.SQLiteDataSource`
    :param selected_record_callback:
        Callback to call when a record is selected in the grid
    :type selected_record_callback: `callable`
    :param selected_record_callback:
        Callback to call when an icon is activated on `DataGridIconView`
    :type selected_record_callback: `callable`
    :param bool has_checkboxes: Whether record rows have a checkbox
    :param decode_fallback: Optional callable for converting objects to
        strings in case `unicode(obj)` fails.
    :type decode_fallback: callable
    :param get_full_path: Callable for returning full paths to files when
        given a relative path, or None if the file isn't available.
    :type get_full_path: callable

    """

    def __init__(self, container, data_source, selected_record_callback=None,
                 activated_icon_callback=None, activated_row_callback=None,
                 has_checkboxes=True, decode_fallback=None,
                 get_full_path=None):
        """Setup UI controls and load initial data view."""
        self.extra_filter_widgets = {}
        self.container = container

        self.decode_fallback = decode_fallback if decode_fallback else repr
        self.get_full_path = get_full_path
        self.selected_record_callback = selected_record_callback
        self.activated_icon_callback = activated_icon_callback
        self.activated_row_callback = activated_row_callback

        self.vscroll = container.grid_scrolledwindow.get_vadjustment()
        self.vscroll.connect_after('value-changed', self.on_scrolled)

        self.tree_view = DataGridView(None, has_checkboxes=has_checkboxes)
        self.icon_view = DataGridIconView(None, has_checkboxes=has_checkboxes)

        self.tree_view.connect('cursor-changed',
                               self.on_treeview_cursor_changed)
        self.tree_view.connect('row-activated',
                               self.on_treeview_row_activated)
        self.icon_view.connect('selection-changed',
                               self.on_iconview_selection_changed)
        self.icon_view.connect('item-activated',
                               self.on_iconview_item_activated)
        self.tree_view.connect('row-expanded',
                               self.on_tree_view_row_expanded)
        self.tree_view.connect('row-collapsed',
                               self.on_tree_view_row_collapsed)

        # The treview will be the default view
        self.view = self.tree_view
        self.container.grid_scrolledwindow.add(self.view)

        cm = ImageCacheManager.get_default()
        cm.connect('image-loaded', self.on_image_cache_manager_image_loaded)

        # select columns toggle button
        self.options_popup = OptionsPopup(
            self.container.togglebutton_options, self)
        self.options_popup.connect('column-visibility-changed',
                                   self.on_popup_column_visibility_changed)
        self.options_popup.connect('view-changed', self.on_popup_view_changed)

        # date range widgets
        icon_theme = Gtk.IconTheme.get_default()
        for icon in ['calendar', 'stock_calendar']:
            if icon_theme.has_icon(icon):
                break
        else:
            # Should never happen, just a precaution
            raise Exception("No suitable calendar icon found on theme")

        for image in [self.container.image_start_date,
                      self.container.image_end_date]:
            image.set_from_icon_name(icon, Gtk.IconSize.BUTTON)

        self.date_start = DateEntry(self.container.window,
                                    DateEntry.TYPE_START)
        self.date_start.set_editable(False)
        self.date_start.set_sensitive(False)
        self.date_start.connect('date_changed', self.on_date_change, 'start')
        # FIXME: ^^ use hyphen in signal name
        self.container.vbox_start_date.pack_start(
            self.date_start, expand=False, fill=True, padding=0)
        self.date_end = DateEntry(self.container.window,
                                  DateEntry.TYPE_END)
        self.date_end.set_editable(False)
        self.date_end.set_sensitive(False)
        self.date_end.connect('date_changed', self.on_date_change, 'end')
        self.container.vbox_end_date.pack_start(
            self.date_end, expand=False, fill=True, padding=0)

        # search widget
        self.container.entry_search.connect('activate', self.on_search_clicked)
        self.container.entry_search.connect(
            'search-changed', self.on_search_clicked)

        self.container.grid_vbox.show_all()

        self.bind_datasource(data_source)

    ###
    # Public
    ###

    def bind_datasource(self, data_source):
        """Binds a data source to the datagrid.

        :param data_source: The data source to bind.
        :type data_source: :class:`datagrid_gtk3.db.DataSource`
        """
        self.model = DataGridModel(data_source,
                                   self.get_full_path,
                                   self.decode_fallback)
        self.model.connect('data-loaded', self.on_data_loaded)
        for view in [self.tree_view, self.icon_view]:
            view.model = self.model

        liststore_date_cols = Gtk.ListStore(str, str, str)
        if self.model.datetime_columns:
            self.date_start.set_sensitive(True)
            self.date_end.set_sensitive(True)

        for column in self.model.datetime_columns:
            liststore_date_cols.append(
                (column['name'], column['display'], column['transform']))

        combox_date_cols = self.container.combobox_date_columns
        old_model = combox_date_cols.get_model()
        if old_model:
            del old_model
        combox_date_cols.set_model(liststore_date_cols)
        if not combox_date_cols.get_cells():
            cell = Gtk.CellRendererText()
            combox_date_cols.pack_start(cell, True)
            combox_date_cols.add_attribute(cell, 'text', 1)
            combox_date_cols.set_active(0)
            combox_date_cols.connect('changed', self.on_date_change, None)

        # Hide date column selection if there can be no choice
        if len(liststore_date_cols) < 2:
            combox_date_cols.hide()
            self.container.date_column_label.hide()
        else:
            # They might have been hidden on a previous bind call.
            combox_date_cols.show()
            self.container.date_column_label.show()

        # If the are no date columns, hide the date range controls as well
        widgets = (
            self.container.image_start_date,
            self.container.vbox_start_date,
            self.container.label_date_to,
            self.container.image_end_date,
            self.container.vbox_end_date,
            self.container.filters_separator,
        )
        if len(liststore_date_cols) == 0:
            for widget in widgets:
                widget.hide()
        else:
            combox_date_cols.set_active(0)
            for widget in widgets:
                widget.show()

        self._refresh_view()
        data_source.connect('rows-changed', self.on_data_source_rows_changed)

    def add_options_filter(self, attr, options, add_empty_option=True):
        """Add optional options filter for attr.

        :param str attr: the attr that will be filtered
        :param iterable options: the options that will be displayed
            on the filter as a tuple with (label, option).
            The label will be displayed as on the combo and the
            option will be used to generate the WHERE clause
        :param bool add_empty_option: if we should add an empty
            option as the first option in the combo. Its label will
            be the label of the column in question and selecting
            it will be the same as not filtering by that attr.
        :returns: the newly created combobox
        :rtype: :class:`Gtk.ComboBox`
        """
        for col_dict in self.model.columns:
            if col_dict['name'] == attr:
                label = col_dict['display']
                break
        else:
            raise ValueError

        model = Gtk.ListStore(str, object)
        if add_empty_option:
            model.append((label, NO_FILTER_OPTION))
        for option in options:
            model.append(option)

        combo = Gtk.ComboBox()
        combo.set_model(model)
        renderer = Gtk.CellRendererText()
        combo.pack_start(renderer, True)
        combo.add_attribute(renderer, 'text', 0)
        combo.set_active(0)
        combo.set_id_column(0)

        combo.connect('changed', self.on_filter_changed, attr)

        self.extra_filter_widgets[attr] = combo
        self.container.extra_filters.pack_start(
            combo, expand=False, fill=False, padding=0)
        self.container.extra_filters.show_all()
        # Make sure this separator is visible. It may have been hidden
        # if we don't have any datetime columns.
        self.container.filters_separator.show()

        return combo

    def set_selected_row_by_id(self, row_id,
                               scroll_to_row=True, callback=None):
        """Set the selected row by id.

        :param int row_id: The id of the row to set as the selected one
        :param bool scroll_to_row: If we should scroll to that row,
            putting it in the center of the window
        :param callable callback: A callback to call after the
            rows has been loaded
        """
        row = self.model.get_row_by_id(row_id, load_rows=True)
        if row is None:
            logger.warning("No row found with id %s", row_id)
            if callback is not None:
                assert callable(callback)
                callback()
            return

        path = Gtk.TreePath(row.path)
        self.view.set_cursor(path)


        def _callback():
            """Timeout callback."""
            if scroll_to_row:
                with self.view.block_draw():
                    if self.view is self.tree_view:
                        self.view.scroll_to_cell(path, None, True, 0.5, 0.0)
                    elif self.view is self.icon_view:
                        self.view.scroll_to_path(path, True, 0.5, 0.0)

            if callback is not None:
                assert callable(callback)
                GObject.idle_add(callback)

        # Acording to the documentation, PRIORITY_HIGH_IDLE + 20 is
        # used by redrawing operations so PRIORITY_HIGH_IDLE + 25
        # should be enought to make sure we call callback just after
        # the widget finishes redrawing itself.
        GObject.timeout_add(100, _callback,
                            priority=GLib.PRIORITY_HIGH_IDLE + 25)

    ###
    # Callbacks
    ###

    def on_scrolled(self, vadj):
        """Load new records upon scroll to end of visible rows.

        :param vadj: Adjustment widget associated with vertical scrollbar
        :type vadj: :class:`Gtk.Adjustment`
        """
        scrolled_to_bottom = (
            vadj.get_value() == (vadj.get_upper() - vadj.get_page_size()) or
            vadj.get_page_size() == vadj.get_upper())

        if scrolled_to_bottom:
            self.model.add_rows()

        self._set_visible_range()

        return False

    def on_popup_column_visibility_changed(self, popup, name, value):
        """Set the list of columns to display based on column checkboxes.

        :param popup: the columns popup
        :type popup: :class:`OptionsPopup`
        :param str name: the name of the columns
        : param bool value: the new column visibility
        """
        if value:
            self.model.display_columns.add(name)
        else:
            self.model.display_columns.discard(name)

        self.model.data_source.set_visible_columns(
            self.model.display_columns)
        self.view.refresh()

    def on_popup_view_changed(self, popup, new_view):
        """Set the actual view based on the options popup option.

        :param popup: the columns popup
        :type popup: :class:`OptionsPopup`
        :param int new_view: either :attr:`OptionsPopup.VIEW_TREE` or
            :attr:`OptionsPopup.VIEW_ICON`
        """
        if new_view == OptionsPopup.VIEW_ICON:
            self.view = self.icon_view
            self.tree_view.set_model(None)
            self.model.image_max_size = 100.0
            self.model.image_draw_border = True
        elif new_view in [OptionsPopup.VIEW_TREE, OptionsPopup.VIEW_FLAT]:
            # Changing view from/to flat will make expanded_ids have no meaning
            self.tree_view.expanded_ids.clear()
            self.view = self.tree_view
            self.icon_view.set_model(None)
            self.model.image_max_size = 24.0
            self.model.image_draw_border = False
        else:
            raise AssertionError("Unrecognized option %r" % (new_view, ))

        # We want flat for flat and iconview, and only if we really have
        # a flat column.
        self.model.active_params['flat'] = (
            self.model.flat_column_idx is not None and
            new_view in [OptionsPopup.VIEW_FLAT, OptionsPopup.VIEW_ICON])

        child = self.container.grid_scrolledwindow.get_child()
        self.container.grid_scrolledwindow.remove(child)
        self.container.grid_scrolledwindow.add(self.view)
        self.view.show_all()

        self._refresh_view()
        # FIXME: Is there a way to keep the selection after the view was
        # refreshed? The actual selected paths are not guaranteed to be the
        # same, so how can we get them again?
        if self.selected_record_callback:
            self.selected_record_callback(None)

    def on_treeview_cursor_changed(self, view):
        """Get the data for a selected record and run optional callback.

        :param view: The treeview containing the row
        :type view: Gtk.TreeView

        """
        selection = view.get_selection()
        model, row_iterator = selection.get_selected()
        if row_iterator and self.selected_record_callback:
            record = self.model.data_source.get_single_record(
                model[row_iterator][self.model.id_column_idx])
            self.selected_record_callback(record)
        elif self.selected_record_callback:
            self.selected_record_callback(None)

    def on_iconview_selection_changed(self, view):
        """Get the data for a selected record and run optional callback.

        :param view: The icon view containing the selected record
        :type view: :class:`Gtk.IconView`
        """
        selections = view.get_selected_items()
        row_iterator = selections and self.model.get_iter(selections[0])
        if row_iterator and self.selected_record_callback:
            model = view.get_model()
            record = self.model.data_source.get_single_record(
                model[row_iterator][self.model.id_column_idx])
            self.selected_record_callback(record)
        elif self.selected_record_callback:
            self.selected_record_callback(None)

    def on_iconview_item_activated(self, view, path):
        """Get the data the activated record and run optional callback.

        :param view: The icon view containing the selected record
        :type view: :class:`Gtk.IconView`
        :param path: the activated path
        """
        if not path or not self.activated_icon_callback:
            return

        row_iterator = view.model.get_iter(path)
        record = self.model.data_source.get_single_record(
            self.model[row_iterator][self.model.id_column_idx])
        self.activated_icon_callback(record, view.pixbuf_column)

    def on_treeview_row_activated(self, view, path, column):
        """Handle row-activated signal on the treeview.

        Run the optional :obj:`.activated_row_callback` when
        a row gets activated.

        :param view: The treeview containing the row
        :type view: :class:`Gtk.TreeView`
        :param path: the activated path
        :type path: :class:`Gtk.TreePath`
        :param column: the column that was activated on the row
        :type column: class:`Gtk.TreeViewColumn`

        """
        if self.activated_row_callback is None:
            return

        row = self.model[self.model.get_iter(path)]
        selected_id = row[self.model.id_column_idx]
        record = self.model.data_source.get_single_record(selected_id)
        self.activated_row_callback(record)

    def on_tree_view_row_expanded(self, treeview, iter_, path):
        """Handle row-expanded events.

        Make sure visible range will be updated on the model.

        :param treeview: the treeview that had one of its rows expanded
        :type treeview: :class:`Gtk.TreeView`
        :param iter_: the iter pointing to the expanded row
        :type iter_: :class:`Gtk.TreeIter`
        :param path: the path pointing to the expanded row
        :type path: :class:`Gtk.TreePath`
        """
        GObject.idle_add(self._set_visible_range)

    def on_tree_view_row_collapsed(self, treeview, iter_, path):
        """Handle row-collapsed events.

        Make sure visible range will be updated on the model.

        :param treeview: the treeview that had one of its rows collapsed
        :type treeview: :class:`Gtk.TreeView`
        :param iter_: the iter pointing to the collapsed row
        :type iter_: :class:`Gtk.TreeIter`
        :param path: the path pointing to the collapsed row
        :type path: :class:`Gtk.TreePath`
        """
        GObject.idle_add(self._set_visible_range)

    def on_image_cache_manager_image_loaded(self, cm):
        """Handle image-loaded event for image cache manager.

        When an image finishes loading, queue a redraw to make sure
        the image will be loaded.

        :param cm: the cache manager that emited the event
        :type cm: :class: `datagrid_gtk3.utils.imageutils.ImageCacheManager`
        """
        self.view.queue_draw()

    def on_filter_changed(self, combo, attr):
        """Handle selection changed on filter comboboxes.

        :param combo: the combo that received the signal
        :type combo: :class:`Gtk.ComboBox`
        :param str attr: the name of the attr to filter
        """
        model = combo.get_model()
        value = model[combo.get_active()][1]

        if value is NO_FILTER_OPTION:
            remove_keys = [attr]
            update_dict = None
        else:
            remove_keys = None
            update_dict = {
                attr: {
                    'operator': 'is' if value is None else '=',
                    'param': value,
                }
            }

        self._refresh_view(update_dict=update_dict, remove_keys=remove_keys)

    def on_data_source_rows_changed(self, data_source, params, ids):
        """Handle data_source rows-changed signal.

        When a row gets updated on data source, make sure to reflect
        them on the view, without requiring to do an extra query for that.

        :param data_source: The data source which emitted the signal
        :type data_source: `datagrid_gtk3.db.DataSource`
        :param params: A dict of params that got updated, mapped as
            ``column_name: new_value``
        :type params: dict
        :param ids: The row ids that got updated
        :type ids: [int]
        """
        params_idx = [
            (self.model.data_source.columns_idx[k], v)
            for k, v in params.iteritems()]
        rows = (row
                for id_, row in self.model.row_id_mapper.iteritems()
                if ids is None or id_ in ids)

        for row in rows:
            for idx, value in params_idx:
                row.data[idx] = value

            path = Gtk.TreePath(row.path)
            self.model.row_changed(path, self.model.get_iter(row.path))
            # Even if we call `view.queue_draw` here, it would only be updated
            # when it got focused. By setting refresh_draw to True, it will
            # force it to refresh the values when the view gets visible on the
            # screen, even if it is not focused atm.
            self.view.refresh_draw = True

    def on_data_loaded(self, model, total_recs):
        """Update the total records label.

        :param model: Current datagrid model
        :type model: :class:`DataGridModel`
        :param int total_recs: Total records for current query

        """
        self.container.label_num_recs.set_markup(
            '<small>%d records</small>' % total_recs
        )

    def on_search_clicked(self, widget):
        """Execute the full-text search for given keyword.

        :param widget: The widget that called the event
        :type widget: :class:`Gtk.Widget`
        """
        search = self.container.entry_search.get_text()
        update_dict = {
            'search': {
                'operator': '=',
                'param': search
            }
        }
        self._refresh_view(update_dict)

    def on_date_change(self, widget, data=None):
        """Refresh the view with chosen date range.

        :param widget: The widget that called the event
        :type widget: :class:`Gtk.Widget`
        :param data: Arbitrary data passed by widget.
        :data type: None
        """
        initial = datetime.datetime(1970, 1, 1)

        start_date = self.date_start.get_date()
        start_timestamp = start_date and (start_date - initial).total_seconds()
        end_date = self.date_end.get_date()
        end_timestamp = end_date and (end_date - initial).total_seconds()

        active_date_column = self.container.combobox_date_columns.get_active()
        model_date_columns = self.container.combobox_date_columns.get_model()
        # clear all params from previous date column range select
        remove_columns = [column[0] for column in model_date_columns]
        # FIXME: Why this is comming as -1 on exampledata for Employee?
        active_date_column = max(active_date_column, 0)
        column = model_date_columns[active_date_column][0]
        transform = model_date_columns[active_date_column][2]

        # Normalize the timestamp so the search is accurate
        if start_timestamp:
            start_timestamp = normalize_timestamp(
                start_timestamp, transform, inverse=True)
        if end_timestamp:
            end_timestamp = normalize_timestamp(
                end_timestamp, transform, inverse=True)

        if start_timestamp is not None and end_timestamp is not None:
            operator = 'range'
            param = (start_timestamp, end_timestamp)
        elif start_timestamp is not None:
            operator = '>='
            param = start_timestamp
        elif end_timestamp is not None:
            operator = '<='
            param = end_timestamp
        else:
            operator = None
            param = None

        update_dict = {}
        if operator is not None:
            update_dict[column] = {
                'operator': operator,
                'param': param,
            }

        self._refresh_view(update_dict, remove_columns)

    ###
    # Private
    ###

    def _refresh_view(self, update_dict=None, remove_keys=None):
        """Reload the grid with any filter/sort parameters.

        :param dict update_dict: Any ``where`` parameters with which to update
            the currently active parameters
        :param remove_keys: List of keys to delete from ``where`` parameters
        :type remove_keys: list

        """
        where_dict = self.model.active_params.setdefault('where', {})

        if remove_keys:
            for key in remove_keys:
                where_dict.pop(key, None)

        where_dict.update(update_dict or {})

        # If in the end the dict was/became empty, remove it from active_params
        if not where_dict:
            del self.model.active_params['where']

        # This will speed up loading the views. Images will be loaded
        # after when _set_visible_range executes.
        self.model.visible_range = ((-1, ), (-1, ))
        self.view.refresh()
        GObject.idle_add(self._set_visible_range)

    def _set_visible_range(self):
        """Update model with current view's visible range."""
        visible_range = self.view.get_visible_range()
        if visible_range is None:
            return

        self.model.visible_range = (
            tuple(visible_range[0]), tuple(visible_range[1]))

        self.view.queue_draw()


class DataGridView(Gtk.TreeView):

    """A ``Gtk.TreeView`` for displaying data from a ``DataGridModel``.

    :param model: The model providing the tabular data for the grid
    :type model: :class:`DataGridModel`
    :keyword bool has_checkboxes: Whether record rows have a checkbox

    """

    has_checkboxes = GObject.property(type=bool, default=True)

    # Column widths
    MIN_WIDTH = 40
    MAX_WIDTH = 400
    SAMPLE_SIZE = 50

    def __init__(self, model, **kwargs):
        super(DataGridView, self).__init__(**kwargs)

        self.connect_after('notify::model', self.after_notify_model)
        self.connect('row-expanded', self.on_row_expanded)
        self.connect('row-collapsed', self.on_row_collapsed)

        # FIXME: Ideally, we should pass model directly to treeview and get
        # it from self.get_model instead of here. We would need to refresh
        # it first though
        self.model = model
        self.check_btn_toggle_all = None
        self.check_btn_toggled_id = None
        self.set_rules_hint(True)
        self.active_sort_column = None
        self.active_sort_column_order = None

        self.expanded_ids = set()
        self._all_expanded = False
        self._block_all_expanded = False

        self.refresh_draw = False
        self._block_draw = False

    ###
    # Public
    ###

    @contextlib.contextmanager
    def block_draw(self):
        """Block the drawing of this widget.

        While inside this context, this widget will not be draw.
        """
        self._block_draw = True
        yield
        self._block_draw = False
        self.queue_draw()

    def do_draw(self, cr):
        """Do the drawing of this widget.

        :param cr: The cairo context context
        :type cr: :class:`cairo.Context`
        """
        if self._block_draw:
            return
        if self.refresh_draw:
            GObject.idle_add(self.queue_draw)
            self.refresh_draw = False
        return Gtk.TreeView.do_draw(self, cr)

    def refresh(self):
        """Refresh the model results."""
        self.set_model(None)
        self.model.refresh()
        self.set_model(self.model)

        for col in self.get_columns()[:]:
            self.remove_column(col)

        self._setup_columns()

        # After refreshing the model, some rows may not be present anymore.
        # Let self.expanded_ids be constructed again by the events bellow
        expanded_ids = self.expanded_ids.copy()
        self.expanded_ids.clear()

        # FIXME: This is very optimized, but on some situations (e.g. all paths
        # are expanded and the user changed the sort column), this would make
        # everything be loaded at the same time. Is there anything we can do
        # regarding this issue?
        for row_id in expanded_ids:
            row = self.model.get_row_by_id(row_id, load_rows=True)
            if row is None:
                continue
            self.expand_to_path(Gtk.TreePath(row.path))

    ###
    # Callbacks
    ###

    def after_notify_model(self, treeview, p_spec):
        """Track model modification on the treeview.

        Aftwe the model of this treeview has changed, we need
        to update some connections, like the 'row-changed'

        :param treeview: the treeview that had its model modified
        :type treeview: `Gtk.TreeView`
        """
        model = treeview.get_model()
        if model is None:
            return

        model.connect('row-changed', self.on_model_row_changed)

    def on_row_expanded(self, treeview, iter_, path):
        """Handle row-expanded events.

        Keep track of which rows are currently expanded

        :param treeview: the treeview that had one of its rows expanded
        :type treeview: :class:`Gtk.TreeView`
        :param iter_: the iter pointing to the expanded row
        :type iter_: class:`Gtk.TreeIter`
        :param path: the path pointing to the expanded row
        :type path: :class:`Gtk.TreePath`
        """
        row_id = self.model.get_value(iter_, self.model.id_column_idx)
        self.expanded_ids.add(row_id)

    def on_row_collapsed(self, treeview, iter_, path):
        """Handle row-collapsed events.

        Keep track of which rows are currently expanded

        :param treeview: the treeview that had one of its rows collapsed
        :type treeview: :class:`Gtk.TreeView`
        :param iter_: the iter pointing to the collapsed row
        :type iter_: class:`Gtk.TreeIter`
        :param path: the path pointing to the collapsed row
        :type path: :class:`Gtk.TreePath`
        """
        self.expanded_ids.discard(
            self.model.get_value(iter_, self.model.id_column_idx))

    def on_model_row_changed(self, model, path, iter_):
        """Track row changes on model.

        :param model: this treeview's model
        :type model: :class:`DataGridModel`
        :param path: the path to the changed row
        :type path: :class:`Gtk.TreePath`
        :param iter_: the iter to the changed row
        :type iter_: :class:`Gtk.TreeIter`
        """
        self._update_toggle_check_btn_activity()

    def on_toggle(self, cell, path, col_index):
        """Toggle row selected checkbox, and update the model.

        :param cell: The toggle renderer widget
        :type cell: :class:`Gtk.CellRendererToggle`
        :param int path: int representing the row in the view
        :param int col_index: The column the toggle widget is in

        """
        if path is not None:
            itr = self.model.get_iter(path)
            val = self.model.get_value(itr, col_index)
            self.model.set_value(itr, col_index, not val)

    def on_tvcol_clicked(self, widget, column):
        """Sort the records by the given column.

        :param widget: The widget of the column being sorted
        :type widget: :class:`Gtk.TreeViewColumn`
        :param column: The column name being sorted, used for query construct
        :type column: str

        """
        sort_order = widget.get_sort_order()
        for col in self.get_columns():
            # remove sort indicators from inactive cols
            col.set_sort_indicator(False)
        widget.set_sort_indicator(True)
        if sort_order == Gtk.SortType.ASCENDING:
            new_sort_order = Gtk.SortType.DESCENDING
        else:
            new_sort_order = Gtk.SortType.ASCENDING
        widget.set_sort_order(new_sort_order)
        self.active_sort_column = column
        self.active_sort_column_order = new_sort_order
        desc = sort_order == Gtk.SortType.DESCENDING
        self.model.active_params.update({'order_by': column, 'desc': desc})
        self.refresh()

    def on_select_all_column_clicked(self, check_btn):
        """Select all records in current recordset and update model/view.

        :param check_btn: The check button inside the treeview column header
        :type: :class:`Gtk.CheckButton`

        """
        val = check_btn.get_active()

        where_params = {}
        if 'where' in self.model.active_params:
            where_params['where'] = self.model.active_params['where']

        ids = self.model.data_source.get_all_record_ids(where_params)
        self.model.update_data_source(
            self.model.data_source.SELECTED_COLUMN, val, ids)

        self.refresh()

    ###
    # Private
    ###

    def _setup_columns(self):
        """Configure the column widgets in the view."""
        if self.has_checkboxes:
            toggle_cell = Gtk.CellRendererToggle()
            toggle_cell.connect('toggled', self.on_toggle,
                                self.model.data_source.selected_column_idx)
            col = Gtk.TreeViewColumn(
                '', toggle_cell,
                active=self.model.data_source.selected_column_idx)

            check_btn = Gtk.CheckButton()
            col.set_widget(check_btn)
            check_btn.show()

            self.check_btn_toggled_id = check_btn.connect(
                "toggled", self.on_select_all_column_clicked)

            # Mimic toggle on checkbutton since it won't receive the click.
            # This will work when clicking directly on the checkbutton or on
            # the header button itself.
            col.connect(
                'clicked',
                lambda tvc: check_btn.set_active(not check_btn.get_active()))

            self.check_btn_toggle_all = check_btn
            self.append_column(col)

        hidden_columns = self.model.hidden_columns
        samples = list(itertools.islice(
            (r.data for r in self.model.iter_rows()), self.SAMPLE_SIZE))
        for column_index, column in enumerate(self.model.columns):
            item = column['name']
            display = item in self.model.display_columns
            if display and column['name'] not in hidden_columns:
                item_display = column['display']
                if column['transform'] in ['boolean', 'image']:
                    renderer = Gtk.CellRendererPixbuf()
                    cell_renderer_kwargs = {'pixbuf': column_index}
                else:
                    renderer = Gtk.CellRendererText()
                    renderer.set_property('ellipsize', Pango.EllipsizeMode.END)
                    if column['type'] in (int, long, float):
                        renderer.set_property('xalign', 1)
                    cell_renderer_kwargs = {'text': column_index}
                lbl = '%s' % (item_display.replace('_', '__'),)
                col = Gtk.TreeViewColumn(lbl, renderer, **cell_renderer_kwargs)
                col.connect('clicked', self.on_tvcol_clicked, item)
                col.set_resizable(True)
                # Set the minimum width for the column based on the width
                # of the label and some padding
                width = self._get_pango_string_width(lbl) + 14
                col.set_fixed_width(
                    self._get_best_column_width(column_index, samples))
                col.set_sizing(Gtk.TreeViewColumnSizing.FIXED)
                col.set_expand(column['expand'])
                if item == self.active_sort_column:
                    # When the column is expanded, leave a little more
                    # space for the sort indicator
                    width += self._get_pango_string_width(
                        u" \u25BC".encode('utf-8'))
                    col.set_sort_indicator(True)
                    col.set_sort_order(self.active_sort_column_order)
                col.set_min_width(width)
                self.append_column(col)

        self.set_headers_clickable(True)
        self._update_toggle_check_btn_activity()

    def _update_toggle_check_btn_activity(self):
        """Update the "selected" treeview column's checkbox activity.

        This will update the checkbox activity based on the selected
        rows on the model.
        """
        if self.check_btn_toggle_all is None:
            return

        selected_idx = self.model.data_source.selected_column_idx
        all_selected = all(row[selected_idx] for row in self.model)
        any_selected = any(row[selected_idx] for row in self.model)

        with self.check_btn_toggle_all.handler_block(
                self.check_btn_toggled_id):
            self.check_btn_toggle_all.set_active(all_selected)
            self.check_btn_toggle_all.set_inconsistent(
                not all_selected and any_selected)

    @staticmethod
    def _get_pango_string_width(string):
        """Get the width of a string in pixels.

        Based on:
        http://python.6.x6.nabble.com/Getting-string-with-in-pixels-td1944346.html

        :param string: String to be measured.
        :return: Width of the string in pixels using the default text font.
        :rtype: int
        """
        label = Gtk.Label()
        pango_layout = label.get_layout()
        pango_layout.set_markup(string)
        pango_layout.set_font_description(label.get_style().font_desc)
        width, _ = pango_layout.get_pixel_size()
        label.destroy()
        return width

    def _get_best_column_width(self, colnum, samples):
        """Determine a reasonable column width for the given column.

        :param int colnum: Index of column
        :param int samples: Number of rows to use to determine best width
        :return: optimal column width
        :rtype: int
        """
        label = '  %s  ' % self.model.columns[colnum]['display']
        layout = self.create_pango_layout(label)
        label_width = layout.get_pixel_size()[0]
        lengths = set()
        model = self.get_model()
        for row in samples:
            value = model.get_formatted_value(row[colnum], colnum,
                                              visible=False)
            if isinstance(value, basestring):
                lines = value.splitlines()
                if lines:
                    value = lines[0]
                del lines
                try:
                    layout = self.create_pango_layout('  %s  ' % value)
                except TypeError:
                    # NOTE: unescaped hex data can cause an error like this:
                    # TypeError: Gtk.Widget.create_pango_layout() argument 1
                    #   must be string without null bytes, not unicode
                    layout = self.create_pango_layout('')
                lengths.add(layout.get_pixel_size()[0])
        if lengths:
            max_length = max(lengths)
        else:
            max_length = 1
        width = max_length
        if width < self.MIN_WIDTH:
            width = self.MIN_WIDTH
            if width < label_width:
                width = label_width + 20
        elif width > self.MAX_WIDTH:
            width = self.MAX_WIDTH
        return width


class DataGridCellAreaRenderer(Gtk.CellAreaBox):

    """A cell area renderer with a check box in it."""

    # 16 is the default size, as defined here:
    # https://git.gnome.org/browse/gtk+/tree/gtk/gtkcheckbutton.c
    CHECK_BUTTON_DEFAULT_SIZE = 16
    CHECK_BUTTON_OFFSET = 6

    def __init__(self, *args, **kwargs):
        super(DataGridCellAreaRenderer, self).__init__(*args, **kwargs)

        checkbutton = Gtk.CheckButton()
        value = GObject.Value(GObject.TYPE_INT)
        checkbutton.style_get_property('indicator-size', value)
        self._checkbutton_size = (value.get_int() or
                                  self.CHECK_BUTTON_DEFAULT_SIZE)

        self._is_checked = False

    ###
    # Public
    ###

    def get_checkbutton_area(self, cell_area):
        """Get the area to draw the checkbox on the cell area.

        :param cell_area: the cell area rectangle
        :type cell_area: `cairo.Rectangle`
        :returns: a tuple with the area as (x, y, width, height)
        :rtype: `tuple`
        """
        pos_x = cell_area.x + self.CHECK_BUTTON_OFFSET
        pos_y = (cell_area.y + cell_area.height -
                 (self._checkbutton_size + self.CHECK_BUTTON_OFFSET))
        return pos_x, pos_y, self._checkbutton_size, self._checkbutton_size

    ###
    # Virtual overrides
    ###

    def do_render(self, ctx, widget, cr, background_area,
                  cell_area, flags, paint_focus):
        """Render the checkbox on the cell area.

        :param ctx: the cell area context
        :type ctx: `Gtk.CellAreaContext`
        :param widget: the widget that we are rendering on
        :type widget: `DataGridIconView`
        :param cr: the context to render with
        :type cr: `cairo.Context`
        :param background_area: the widget relative coordinates from
            the area's background
        :type background_area: `cairo.Rectangle`
        :param cell_area: the widget relative coordinates from area
        :type cell_area: `cairo.Rectangle`
        :param flags: the cell renderer state for the area in this row
        :type flags: `Gtk.CellRendererState`
        :param paint_focus: wheather the area should paint focus on
            focused cells for focused rows or not
        :type paint_focus: `bool`
        """
        # For some reason, can't use super here
        Gtk.CellAreaBox.do_render(
            self, ctx, widget, cr, background_area,
            cell_area, flags, paint_focus)

        if not widget.has_checkboxes:
            return

        style_context = widget.get_style_context()
        style_context.save()
        style_context.add_class(Gtk.STYLE_CLASS_CHECK)

        if self._is_checked:
            # CHECKED is the right flag to use, but it's only available on
            # gtk 3.14+. For older versions, we use ACTIVE alone, since
            # setting it together with existing flags would break its drawing
            if hasattr(Gtk.StateFlags, 'CHECKED'):
                style_context.set_state(style_context.get_state() |
                                        Gtk.StateFlags.CHECKED)
            else:
                style_context.set_state(Gtk.StateFlags.ACTIVE)

        Gtk.render_check(widget.get_style_context(), cr,
                         *self.get_checkbutton_area(cell_area))

        style_context.restore()

    def do_apply_attributes(self, model, iter_, *args):
        """Render the checkbox on the cell area.

        :param model: the model to pull values from
        :type model: `DataGridModel`
        :param iter_: the iter to apply values from
        :type iter_: `Gtk.TreeIter`
        """
        # For some reason, can't use super here
        Gtk.CellAreaBox.do_apply_attributes(self, model, iter_, *args)
        if model.data_source.selected_column_idx:
            self._is_checked = model.get_value(
                iter_, model.data_source.selected_column_idx)


class DataGridIconView(Gtk.IconView):

    """A ``Gtk.IconView`` for displaying data from a ``DataGridModel``.

    :param model: The model providing the tabular data for the grid
    :type model: :class:`DataGridModel`
    :keyword bool has_checkboxes: Whether record rows have a checkbox

    """

    has_checkboxes = GObject.property(type=bool, default=True)

    def __init__(self, model, **kwargs):
        if 'cell_area' not in kwargs:
            kwargs['cell_area'] = DataGridCellAreaRenderer()

        super(DataGridIconView, self).__init__(**kwargs)

        self._button_press_time = 0
        self._button_press_scroll = None
        self._button_press_path = None

        self.pixbuf_column = None
        # FIXME: Ideally, we should pass model directly to treeview and get
        # it from self.get_model instead of here. We would need to refresh
        # it first though
        self.model = model

        self.connect('button-release-event', self.on_button_release_event)
        self.connect('key-press-event', self.on_key_press_event)

        self.refresh_draw = False
        self._block_draw = False

    ###
    # Public
    ###

    @contextlib.contextmanager
    def block_draw(self):
        """Block the drawing of this widget.

        While inside this context, this widget will not be draw.
        """
        self._block_draw = True
        yield
        self._block_draw = False
        self.queue_draw()

    def do_draw(self, cr):
        """Do the drawing of this widget.

        :param cr: The cairo context context
        :type cr: :class:`cairo.Context`
        """
        if self._block_draw:
            return
        if self.refresh_draw:
            GObject.idle_add(self.queue_draw)
            self.refresh_draw = False
        return Gtk.IconView.do_draw(self, cr)

    def refresh(self):
        """Refresh the model results."""
        self.set_model(None)
        self.model.refresh()
        self.set_model(self.model)

        if self.model.flat_column_idx is not None:
            # When defining a text column, we need to set a max width or
            # it will expand the rows to fit the longest string.
            self.set_item_width(100)
            self.set_text_column(self.model.flat_column_idx)

        for column_index, column in enumerate(self.model.columns):
            # FIXME: Can we have more than one column with image transform?
            if column['transform'] == 'image':
                self.pixbuf_column = column_index
                self.set_pixbuf_column(self.pixbuf_column)
                break

    ##
    # Callbacks
    ##

    def on_key_press_event(self, window, event):
        """Handle key press events.

        Toggle the check button when pressing 'Space'
        """
        # We don't want 'item-activated' signal to be fired on Space, even
        # if we don't have checkboxes visible. Space should be used
        # to toggle the checkboxes only
        space_pressed = event.get_keyval()[1] == Gdk.KEY_space
        if not space_pressed:
            return False

        if not self.has_checkboxes:
            return space_pressed

        selections = self.get_selected_items()
        if not selections:
            return space_pressed

        self._toggle_path(selections[0])
        return True

    def on_button_release_event(self, window, event):
        """Handle button press events.

        Toggle the check button if we clicked on it.
        """
        coords = event.get_coords()
        path = self.get_path_at_pos(*coords)
        if not path:
            return False

        # If we have checkboxes, check if the click was on it. If it was,
        # we will need to toggle its state.
        if self.has_checkboxes:
            success, cell_rect = self.get_cell_rect(path, None)
            cell_area = self.get_property('cell_area')

            event_rect = Gdk.Rectangle()
            event_rect.x, event_rect.y = coords
            event_rect.width = 1
            event_rect.height = 1

            check_rect = Gdk.Rectangle()
            (x, y,
             check_rect.width,
             check_rect.height) = cell_area.get_checkbutton_area(cell_rect)

            # x and y needs to be converted to bin window coords
            (check_rect.x,
             check_rect.y) = self.convert_widget_to_bin_window_coords(x, y)

            # For some reason, we also need to consider the item padding
            check_rect.x += self.get_item_padding()
            check_rect.y -= self.get_item_padding()

            intersection = Gdk.rectangle_intersect(event_rect, check_rect)
            if intersection[0]:
                self._toggle_path(path)
                return True

        # FIXME: This is to workaround a problem that, if the item's height is
        # greater than the available space (and thus, it is cropped), double
        # clicking it will make the scroll move but not activate it.
        # We check if the scroll is really different to avoid activating the
        # item twice when the item is not cropped.
        # Note: Gtk considers a double click if the first event happened
        # with a difference of a quarter of second from the other.
        event_time = event.get_time()
        scroll = self.get_vadjustment().get_value()
        if (path == self._button_press_path and
                event_time - self._button_press_time <= 250 and
                scroll != self._button_press_scroll):
            self._button_press_time = 0
            self._button_press_scroll = None
            self._button_press_path = None
            self.item_activated(path)
            return True

        self._button_press_scroll = scroll
        self._button_press_time = event_time
        self._button_press_path = path

        return False

    ###
    # Private
    ###

    def _toggle_path(self, path):
        """Toggle the '__selected' value for the given path on model.

        :param path: the path to toggle
        :type itr: :class:`Gtk.TreePath`
        """
        iter_ = self.model.get_iter(path)
        selected_idx = self.model.data_source.selected_column_idx
        val = self.model.get_value(iter_, selected_idx)
        # FIXME: Gtk.IconView has some problems working with huge models.
        # It would invalidate everything on row changed, as can be seem here:
        # https://git.gnome.org/browse/gtk+/tree/gtk/gtkiconview.c
        # Atm we will avoid row-changed and just call queue_redraw which
        # will be a lot faster! We should try to find a better solution
        self.model.set_value(iter_, selected_idx, not val,
                             emit_event=False)
        self.queue_draw()


class DataGridModel(GenericTreeModel):

    """Underlying model for data grid view.

    This is a ``TreeModel`` class for representing data from a persistent data
    store such as a SQLite database table.

    :param data_source: Persistent data source to populate model
    :type data_source: :class:`datagrid_gtk3.db.sqlite.SQLiteDataSource`
    :param get_media_callback: Function to retrieve media file
    :type get_media_callback: callable
    :param decode_fallback: Callable for converting objects to
        strings in case `unicode(obj)` fails.
    :type decode_fallback: callable
    :param str encoding_hint: Encoding to use for rendering strings

    It may be  a question of changing parent class(es) and changing eg.
    ``on_get_flags`` to ``do_get_flags`` etc.

    """

    __gsignals__ = {
        'data-loaded': (GObject.SignalFlags.RUN_FIRST, None, (object,))
    }

    image_max_size = GObject.property(type=float, default=24.0)
    image_draw_border = GObject.property(type=bool, default=False)
    image_load_on_thread = GObject.property(type=bool, default=True)

    STRING_MAX_LENGTH = 100
    IMAGE_PREFIX = 'file://'

    def __init__(self, data_source, get_media_callback, decode_fallback,
                 encoding_hint='utf-8'):
        """Set up model."""
        super(DataGridModel, self).__init__()

        self._invisible_images = {}
        self._fallback_images = {}
        self.visible_range = None
        self.active_params = {'flat': False}
        self.data_source = data_source
        self.get_media_callback = get_media_callback
        self.decode_fallback = decode_fallback
        self.columns = self.data_source.columns
        self.datetime_columns = []
        self.column_types = []
        for column in self.columns:
            transform = column['transform']
            if transform is None:
                continue
            if transform.startswith('timestamp'):
                self.datetime_columns.append(column)
            self.column_types.append(column['type'])

        selected_columns = self.data_source.get_visible_columns()
        if selected_columns is not None:
            self.display_columns = set(selected_columns)
        else:
            self.display_columns = {
                col['name'] for col in self.columns
                if col['visible'] and not col['name'].startswith('__')}

        self.encoding_hint = encoding_hint
        self.selected_cells = list()

        self.row_id_mapper = {}
        self.id_column_idx = None
        self.parent_column_idx = None
        self.flat_column_idx = None
        self.rows = None
        self.total_recs = None

    @property
    def hidden_columns(self):
        """A set of columns names that should not be displayed on the view."""
        hidden = {self.data_source.SELECTED_COLUMN}
        if not self.data_source.display_all:
            hidden.add(self.data_source.ID_COLUMN)
            hidden.add(self.data_source.PARENT_ID_COLUMN)
            if not self.active_params.get('flat', False):
                hidden.add(self.data_source.FLAT_COLUMN)
        return hidden

    def refresh(self):
        """Refresh the model from the data source."""
        if 'page' in self.active_params:
            del self.active_params['page']
        if 'parent_id' in self.active_params:
            del self.active_params['parent_id']

        self.row_id_mapper.clear()
        self.rows = self.data_source.load(self.active_params)
        self.rows.path = ()

        self.id_column_idx = self.data_source.id_column_idx
        self.parent_column_idx = self.data_source.parent_column_idx
        self.flat_column_idx = self.data_source.flat_column_idx
        self.total_recs = self.data_source.total_recs

        if self.id_column_idx is not None:
            for i, row in enumerate(self.rows):
                row.path = (i, )
                self.row_id_mapper[row.data[self.id_column_idx]] = row

        self.emit('data-loaded', self.total_recs)

    def add_rows(self, parent_node=None):
        """Add rows to the model from a new page of data and update the view.

        :return: True if update took place, False if not
        :rtype: bool
        """
        is_tree = (self.parent_column_idx is not None and
                   not self.active_params.get('flat', False))
        # When the data is hierarchical, all the root data was already loaded,
        # except in a flat view, where data is being lazy loaded.
        if is_tree and parent_node is None:
            return False

        if parent_node is None:
            parent_id = None
            parent_row = self.rows
            path_offset = self.rows[-1].path[-1] + 1
            # We are not using pages for hierarchical data
            self.active_params['page'] = self.active_params.get('page', 0) + 1
        else:
            parent_id = parent_node.data[self.id_column_idx]
            parent_row = parent_node
            path_offset = 0

        self.active_params['parent_id'] = parent_id
        rows = self.data_source.load(self.active_params)
        if not len(rows):
            return False

        for i, row in enumerate(rows):
            row.path = parent_row.path + (path_offset + i, )
            self.row_id_mapper[row.data[self.id_column_idx]] = row
            parent_row.append(row)

            # Non-hierarchical data need this to call self.row_inserted
            # so the treemodel will know about the new rows.
            # For hierarchical, those paths/iters already exists, they
            # are just lazy-loaded when neede (i.e. when expanding the parent)
            if not is_tree:
                parent_row.children_len += 1
                path = Gtk.TreePath(row.path)
                self.row_inserted(path, self.create_tree_iter(row.path))

        return True

    def update_data_source(self, column, value, ids):
        """Update the model's persistent data source for given records.

        Currently only used for updating "__selected" column.

        Note that the function that uses this must call ``row_changed`` or
        reset grid in order to see changes.

        :param str column: Name of column to update
        :param value: Update value
        :type value: str or int
        :param list ids: List of primary keys of records to update
        """
        param = {column: value}
        self.data_source.update(param, ids)

    def get_formatted_value(self, value, column_index, visible=True):
        """Get the value to display in the cell.

        :param value: Value from data source
        :type value: str or int or None
        :param int column_index: Index of the column containing the value
        :param bool visible: If the value is visible on the view or not.
            Some transformations (i.e. image) will do some optimizations
            if the value is not visible
        :return: formatted value
        :rtype: unicode or int or bool or None
        """
        col_dict = self.columns[column_index]
        col_name = col_dict['name']

        # Defaults to string transformer if None
        transformer_name = col_dict['transform'] or 'string'
        transformer = get_transformer(transformer_name)
        transformer_kwargs = {}

        # Only enforce value type if the config was provided. Otherwise,
        # we would just be spamming a lot of obvious warnings (we got the type
        # from introspecting the database and for sqlite, it has a high
        # probability of not being an exact match in python).
        if (col_dict['from_config'] and
                value is not None and 'type' in col_dict):
            # Try enforcing value type
            value = self._enforce_value_type(value, col_dict['type'])

        if transformer is None:
            logger.warning("No transformer found for %s", transformer_name)
            return value

        if (transformer_name == 'boolean' and
                col_name == self.data_source.SELECTED_COLUMN):
            # __selected is an exception to the boolean transformation.
            # It requires a bool value and not a pixbuf
            return bool(value)
        elif transformer_name == 'image':
            transformer_kwargs.update(dict(
                size=self.image_max_size,
                draw_border=self.image_draw_border,
                load_on_thread=self.image_load_on_thread,
                draft=True,
            ))

            # If no value, use an invisible image as a placeholder
            if not value:
                invisible_img = self._invisible_images.get(self.image_max_size)
                if not invisible_img:
                    invisible_img = NO_IMAGE_PIXBUF.scale_simple(
                        self.image_max_size, self.image_max_size,
                        GdkPixbuf.InterpType.NEAREST)
                    self._invisible_images[self.image_max_size] = invisible_img
                return invisible_img

            if isinstance(value, buffer):
                # FIXME: Support buffer in the future. Atm, set visible to
                # False so a fallback image will be returned bellow
                logger.warn('Buffered images are still not supported')
                visible = False

            # When not visible on the iconview, use an already generated
            # fallback image (that has the same dimensions as the real
            # image should have) to improve loading time.
            if not visible:
                key = (self.image_draw_border, self.image_max_size)
                fallback = self._fallback_images.get(key)
                if not fallback:
                    fallback = transformer(None, **transformer_kwargs)
                    self._fallback_images[key] = fallback
                return fallback

            if value.startswith(self.IMAGE_PREFIX):
                value = value[len(self.IMAGE_PREFIX):]

            if not value:
                # Force fallback in this case
                value = None
            elif not os.path.isabs(value):
                if self.get_media_callback is None:
                    logger.warning(
                        "Don't know how to access the relative path '%s'. "
                        "Try passing get_full_path to controller.", value)
                    value = None
                else:
                    value = self.get_media_callback(value)
        elif transformer_name in ['string', 'html']:
            transformer_kwargs.update(dict(
                max_length=self.STRING_MAX_LENGTH, oneline=True,
                decode_fallback=self.decode_fallback,
            ))

        custom_options = col_dict.get('transform_options')
        if custom_options:
            transformer_kwargs['options'] = custom_options

        return transformer(value, **transformer_kwargs)

    def set_value(self, itr, column, value, emit_event=True):
        """Set the value in the model and update the data source with it.

        :param itr: ``TreeIter`` object representing the current row
        :type itr: :class:`Gtk.TreeIter`
        :param int column: Column index for value
        :param value: Update the row/column to this value
        :type value: str or int or bool or None
        :param bool emit_event: if we should call :meth:`.row_changed`.
            Be sure to know what you are doing before passind `False` here
        """
        path = self.get_path(itr)
        # path and iter are the same in this model.
        row = self._get_row_by_path(path)
        row.data[column] = value
        id_ = self.get_value(itr, self.id_column_idx)
        self.update_data_source(
            self.columns[column]['name'], value, [int(id_)])
        if emit_event:
            self.row_changed(path, itr)

    def iter_rows(self, load_rows=False):
        """Iterate over the rows of the model.

        This will iterate using a depth-first algorithm. That means that,
        on a hierarchy like this::

            A
              B
                E
              C
                F
                  G
                  H
              D

        This would generate an iteration like::

            [ A, B, E, C, F, G, H, D ]

        :param bool load_rows: if we should load rows from the
            datasource during the iteration. When using this, be sure
            to use lazy iteration and stop when you found what you needed.
        :returns: an iterator for the rows
        :rtype: generator
        """
        def _iter_children_aux(parent):
            for row in parent:
                if load_rows:
                    self._ensure_children_is_loaded(row)
                yield row
                for inner_row in _iter_children_aux(row):
                    yield inner_row

        for row in _iter_children_aux(self.rows):
            yield row

        if load_rows:
            rows_len = len(self.rows)
            while self.add_rows():
                for row in _iter_children_aux(self.rows[rows_len:]):
                    yield row
                rows_len = len(self.rows)

    def get_row_by_id(self, row_id, load_rows=False):
        """Get a row given its id

        Note that this will load the data from the source until the
        row is found, meaning that everything will be loaded on the
        worst case (i.e. the row is not present)

        :param object row_id: the id of the row
        :returns: the row or ``None`` if it wasn't found
        :rtype: :class:`datagrid_gtk3.db.sqlite.Node`
        """
        if row_id in self.row_id_mapper:
            return self.row_id_mapper[row_id]

        for row in self.iter_rows(load_rows=load_rows):
            # Although we could check row, trying self.row_id_mapper has a
            # chance of needing less iterations (and thus, less loading from
            # sqlite) since after loading all children of A, we can find them
            # on self.row_id_mapper without having to load their children too
            if row_id in self.row_id_mapper:
                return self.row_id_mapper[row_id]

    ###
    # Private
    ###

    def _enforce_value_type(self, value, type_):
        # FIXME: Some configurations are indicating the images as buffer,
        # but really are storing the file path. This can be removed
        # when those places are fixed.
        if type_ == buffer and isinstance(value, basestring):
            return value

        # Don't try to convert str=>unicode and unicode=>str
        type_check = basestring if type_ in [str, unicode] else type_
        if not isinstance(value, type_check):
            try:
                value = type_(value)
            except (ValueError, TypeError):
                logger.warn('Could not enforce type "%s" for "%r"',
                            type_.__name__, value)

        return value

    def _ensure_children_is_loaded(self, row):
        if not row.is_children_loaded():
            self.add_rows(row)

    def _get_row_by_path(self, iter_):
        def get_row_by_iter_aux(iter_aux, rows):
            self._ensure_children_is_loaded(rows)
            if len(iter_aux) == 1:
                row = rows[iter_aux[0]]
                return row
            return get_row_by_iter_aux(iter_aux[1:], rows[iter_aux[0]])

        return get_row_by_iter_aux(iter_, self.rows)

    ###
    # Required implementations for GenericTreeModel
    ###

    def on_get_flags(self):
        """Return the GtkTreeModelFlags for this particular type of model."""
        return Gtk.TreeModelFlags.ITERS_PERSIST

    def on_get_n_columns(self):
        """Return the number of columns in the model."""
        return len(self.columns)

    def on_get_column_type(self, index):
        """Return the type of a column in the model."""
        if self.columns[index]["name"] == self.data_source.SELECTED_COLUMN:
            return bool
        else:
            if self.columns[index]['transform'] in ['boolean', 'image']:
                return GdkPixbuf.Pixbuf

            return str
            # NOTE: int/long column types cannot display None/null values
            #   so just use str for everything except pixbufs instead of
            #   self.column_types[index]

    def on_get_path(self, rowref):
        """Return the tree path (a tuple of indices) for a particular node."""
        return tuple(rowref)

    def on_get_iter(self, path):
        """Return the node corresponding to the given path (node is path)."""
        try:
            # row and path are the same in this model. We just need
            # to make sure that the iter is valid
            self._get_row_by_path(path)
        except IndexError:
            return None
        else:
            return tuple(path)

    def on_get_value(self, rowref, column):
        """Return the value stored in a particular column for the node."""
        if self.visible_range:
            visible = self.visible_range[0] <= rowref <= self.visible_range[1]
        else:
            visible = True

        row = self._get_row_by_path(rowref)
        raw = row.data[column]
        # Don't format value for id and parent columns. They are not displayed
        # on the grid and we may need their full values to get their records
        # (e.g. when the id is a string column)
        if column in [self.id_column_idx, self.parent_column_idx]:
            return raw
        else:
            return self.get_formatted_value(raw, column, visible=visible)

    def on_iter_next(self, rowref):
        """Return the next node at this level of the tree."""
        if rowref is None:
            return None

        # root node
        if len(rowref) == 1:
            rows = self.rows
            next_value = (rowref[0] + 1, )
        else:
            parentref = rowref[:-1]
            rows = self._get_row_by_path(parentref)
            next_value = parentref + (rowref[-1] + 1, )

        if not next_value[-1] < rows.children_len:
            return None

        return next_value

    def on_iter_children(self, rowref):
        """Return the first child of this node."""
        if rowref is None:
            return (0, )

        parent_row = self._get_row_by_path(rowref)
        if not parent_row.children_len:
            return None

        return rowref + (0, )

    def on_iter_has_child(self, rowref):
        """Return true if this node has children."""
        row = self._get_row_by_path(rowref)
        return row.children_len > 0

    def on_iter_n_children(self, rowref):
        """Return the number of children of this node."""
        if rowref is None:
            return len(self.rows)

        row = self._get_row_by_path(rowref)
        return row.children_len

    def on_iter_nth_child(self, parent, n):
        """Return the nth child of this node."""
        if parent is None:
            parent = ()
            rows = self.rows
        else:
            rows = self._get_row_by_path(parent)

        if not 0 <= n < rows.children_len:
            return None

        return parent + (n, )

    def on_iter_parent(self, child):
        """Return the parent of this node."""
        if len(child) == 1:
            return None

        return child[:-1]

    ###
    # END Required implementations for GenericTreeModel
    ###



"""Database package.

Modules that provide interaction with persistent data stores via database
backends (eg. SQLite)

"""

from gi.repository import GObject


class Node(list):

    """A list that can hold data.

    Just like a simple list, but one can set/get its data
    from :obj:`.data`.

    :param object data: the data that will be stored in this node
    :param int children_len: the number of the children that will
        be loaded lazely at some point
    """

    def __init__(self, data=None, children_len=0):
        super(Node, self).__init__()

        self.data = data
        self.children_len = children_len
        self.path = None

    def is_children_loaded(self, recursive=False):
        """Check if this node's children is loaded

        :param bool recursive: wheather to ask each child if their
            children is loaded (and their child too and so on) too.
        :returns: `True` if children is loaded, otherwise `False`
        :rtype: bool
        """
        loaded = len(self) == self.children_len
        if recursive:
            loaded = (loaded and
                      all(c.is_children_loaded(recursive=True) for c in self))
        return loaded


class DataSource(GObject.GObject):
    """Base class for data sources."""

    ID_COLUMN = 'rowid'
    SELECTED_COLUMN = '__selected'
    PARENT_ID_COLUMN = None
    CHILDREN_LEN_COLUMN = None
    FLAT_COLUMN = None

    def __init__(self):
        super(DataSource, self).__init__()
        self.columns = []
        self.total_recs = 0
        self.display_all = True
        self.id_column_idx = None
        self.parent_column_idx = None
        self.children_len_column_idx = None
        self.flat_column_idx = None
        self.selected_column_idx = None

    def get_visible_columns(self):
        return []

    def load(self, params=None):
        return Node()

    def update_selected_columns(self, columns):
        pass

    def get_all_record_ids(self, params=None):
        return []

    def get_single_record(self, record_id, table=None):
        return tuple()

    def update(self, params, ids=None):
        pass


class EmptyDataSource(DataSource):
    """Data source that can be used when an empty data grid is required."""

    __gsignals__ = {
        'rows-changed': (GObject.SignalFlags.RUN_LAST, None, (object, object)),
    }



"""Image utilities

Some general image utilities using PIL.
"""

import Queue
import collections
import io
import mimetypes
import os
import struct
import threading

from gi.repository import (
    GLib,
    GObject,
    GdkPixbuf,
    Gtk,
    Gdk,
)
from PIL import Image, ImageFilter

mimetypes.init()
# Generating a drop shadow is an expensive operation. Keep a cache
# of already generated drop shadows so they can be reutilized
_drop_shadows_cache = {}

_icon_theme = Gtk.IconTheme.get_default()
_icon_filename_cache = {}


def get_icon_filename(choose_list, size):
    """Get a theme icon filename.

    :param list choose_list: the list of icon names to choose from.
        The first existing icon will be returned.
    :param int size: size of the icon, to be passed to
        :class:`Gtk.IconTheme.choose_icon`
    :return: the path to the icon
    :rtype: str
    """
    icon = _icon_theme.choose_icon(choose_list, size,
                                   Gtk.IconLookupFlags.NO_SVG)
    return icon and icon.get_filename()


def get_icon_for_file(filename, size):
    """Get icon for filename mimetype.

    Analyze filename to get its mimetype and return the path of an
    icon representing it.

    :param str filename: path of the file to be alalyzed
    :param int size: size of the icon, to be passed to
        :class:`Gtk.IconTheme.choose_icon`
    :return: the path to the icon
    :rtype: str
    """
    if os.path.isdir(filename):
        # mimetypes.guess_type doesn't work for folders
        guessed_mime = 'folder/folder'
    else:
        # Fallback to unknown if mimetypes wasn't able to guess it
        guessed_mime = mimetypes.guess_type(filename)[0] or 'unknown/unknown'

    if guessed_mime in _icon_filename_cache:
        return _icon_filename_cache[guessed_mime]

    # Is there any value returned by guess_type that would have no /?
    mimetype, details = guessed_mime.split('/')

    # FIXME: guess_type mimetype is formatted differently from what
    # Gtk.IconTheme expects. We are trying to improve matching here.
    # Is there a better way for doing this?
    icon_list = ['%s-%s' % (mimetype, details), details, mimetype]
    if mimetype == 'application':
        icon_list.append('application-x-%s' % (details, ))
    icon_list.append('%s-x-generic' % (mimetype, ))
    icon_list.append('unknown')

    icon_filename = get_icon_filename(icon_list, size)
    _icon_filename_cache[guessed_mime] = icon_filename
    return icon_filename


def image2pixbuf(image):
    """Convert a PIL image to a pixbuf.

    :param image: the image to convert
    :type image: `PIL.Image`
    :returns: the newly created pixbuf
    :rtype: `GdkPixbuf.Pixbuf`
    """
    with io.BytesIO() as f:
        image.save(f, 'png')
        loader = GdkPixbuf.PixbufLoader.new_with_type('png')
        loader.write(f.getvalue())
        pixbuf = loader.get_pixbuf()
        loader.close()

    return pixbuf


def add_border(image, border_size=5,
               background_color=(0xff, 0xff, 0xff, 0xff)):
    """Add a border on the image.

    :param image: the image to add the border
    :type image: `PIL.Image`
    :param int border_size: the size of the border
    :param tuple background_color: the color of the border as a
        tuple containing (r, g, b, a) information
    :returns: the new image with the border
    :rtype: `PIL.Image`
    """
    width = image.size[0] + border_size * 2
    height = image.size[1] + border_size * 2

    try:
        image.convert("RGBA")
        image_parts = image.split()
        mask = image_parts[3] if len(image_parts) == 4 else None
    except IOError:
        mask = None

    border = Image.new("RGBA", (width, height), background_color)
    border.paste(image, (border_size, border_size), mask=mask)

    return border


def add_drop_shadow(image, iterations=3, border_size=2, offset=(2, 2),
                    shadow_color=(0x00, 0x00, 0x00, 0xff)):
    """Add a border on the image.

    Based on this receipe::

        http://en.wikibooks.org/wiki/Python_Imaging_Library/Drop_Shadows

    :param image: the image to add the drop shadow
    :type image: `PIL.Image`
    :param int iterations: number of times to apply the blur filter
    :param int border_size: the size of the border to add to leave
        space for the shadow
    :param tuple offset: the offset of the shadow as (x, y)
    :param tuple shadow_color: the color of the shadow as a
        tuple containing (r, g, b, a) information
    :returns: the new image with the drop shadow
    :rtype: `PIL.Image`
    """
    width  = image.size[0] + abs(offset[0]) + 2 * border_size
    height = image.size[1] + abs(offset[1]) + 2 * border_size

    key = (width, height, iterations, border_size, offset, shadow_color)
    existing_shadow = _drop_shadows_cache.get(key)
    if existing_shadow:
        shadow = existing_shadow.copy()
    else:
        shadow = Image.new('RGBA', (width, height),
                           (0xff, 0xff, 0xff, 0x00))

        # Place the shadow, with the required offset
        # if < 0, push the rest of the image right
        shadow_lft = border_size + max(offset[0], 0)
        # if < 0, push the rest of the image down
        shadow_top  = border_size + max(offset[1], 0)

        shadow.paste(shadow_color,
                     [shadow_lft, shadow_top,
                      shadow_lft + image.size[0],
                      shadow_top + image.size[1]])

        # Apply the BLUR filter repeatedly
        for i in range(iterations):
            shadow = shadow.filter(ImageFilter.BLUR)

        _drop_shadows_cache[key] = shadow.copy()

    # Paste the original image on top of the shadow
    # if the shadow offset was < 0, push right
    img_lft = border_size - min(offset[0], 0)
    # if the shadow offset was < 0, push down
    img_top = border_size - min(offset[1], 0)

    shadow.paste(image, (img_lft, img_top))
    return shadow


class ImageCacheManager(GObject.GObject):

    """Helper to cache image transformations.

    Image transformations can be expensive and datagrid views will
    ask for them a lot. This will help by:

        * Caching the mru images so the pixbuf is ready to be used,
          without having to load and transform it again

        * Do the transformations on another thread so larger images
          transformation will not disturb the main one.

    """

    __gsignals__ = {
        'image-loaded': (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    _instance = None

    MAX_CACHE_SIZE = 200
    IMAGE_BORDER_SIZE = 6
    IMAGE_SHADOW_SIZE = 6
    IMAGE_SHADOW_OFFSET = 2

    def __init__(self):
        """Initialize the image cache manager object."""
        super(ImageCacheManager, self).__init__()

        self._lock = threading.Lock()
        self._cache = {}
        self._placeholders = {}
        self._mru = collections.deque([], self.MAX_CACHE_SIZE)
        self._waiting = set()
        # We are using a LifoQueue instead of a Queue to load the most recently
        # used image. For example, when scrolling the treeview, you will want
        # the visible rows to be loaded before the ones that were put in the
        # queue during the process.
        self._queue = Queue.LifoQueue()

        self._task = threading.Thread(target=self._transform_task)
        self._task.daemon = True
        self._task.start()

    ###
    # Public
    ###

    @classmethod
    def get_default(cls):
        """Get the singleton default cache manager.

        :return: the cache manager
        :rtype: :class:`ImageCacheManager`
        """
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def get_image(self, path, size=24, fill_image=True, draw_border=False,
                  draft=False, load_on_thread=False):
        """Render path into a pixbuf.

        :param str path: the image path or `None` to use a fallback image
        :param int size: the size to resize the image. It will be resized
            to fit a square of (size, size)
        :param bool fill_image: if we should fill the image with a transparent
            background to make a smaller image be at least a square of
            (size, size), with the real image at the center.
        :param bool draw_border: if we should add a border on the image
        :param bool draft: if we should load the image as a draft. This
            trades a little quality for a much higher performance.
        :param bool load_on_thread: if we should load the image on another
            thread. This will make a placeholder be returned the first
            time this method is called.
        :returns: the resized pixbuf
        :rtype: :class:`GdkPixbuf.Pixbuf`
        """
        params = (path, size, fill_image, draw_border, draft)

        with self._lock:
            # We want this params to be the last element on the deque (meaning
            # it will be the most recently used item). Since we will append it
            # bellow, if it were already in the deque, make sure to remove it
            if params in self._mru:
                self._mru.remove(params)
            # When self._mru reaches its maxlen, the least recently used
            #element (the position 0 in case of an append) will be removed
            self._mru.append(params)

            pixbuf = self._cache.get(params, None)
            # The pixbuf is on cache
            if pixbuf is not None:
                return pixbuf

            # The pixbuf is not on cache, but we don't want to
            # load it on a thread
            if not load_on_thread:
                pixbuf = self._transform_image(*params)
                # If no pixbuf, let the fallback image be returned
                if pixbuf:
                    self._cache_pixbuf(params, pixbuf)
                    return pixbuf
            elif params not in self._waiting:
                self._waiting.add(params)
                self._queue.put(params)

        # Size will always be rounded to the next value. After 48, the
        # next is 256 and we don't want something that big here.
        fallback_size = min(size, 48)
        fallback = get_icon_for_file(path or '', fallback_size)

        placeholder_key = (fallback, ) + tuple(params[1:])
        placeholder = self._placeholders.get(placeholder_key, None)
        if placeholder is None:
            # If the image is damaged for some reason, use fallback for
            # its mimetype. Maybe the image is not really an image
            # (it could be a video, a plain text file, etc)
            placeholder = self._transform_image(
                fallback, fallback_size, *params[2:])
            self._placeholders[placeholder_key] = placeholder
            # Make the placeholder the initial value for the image. If the
            # loading fails, it will be used as the pixbuf for the image.
            self._cache[params] = placeholder

        return placeholder

    ###
    # Private
    ###

    def _cache_pixbuf(self, params, pixbuf):
        """Cache the pixbuf.

        Cache the pixbuf generated by the given params.
        This will also free any item any item that is not needed
        anymore (the least recently used items after the
        cache > :attr:`.MAX_CACHE_SIZE`) from the cache.

        :param tuple params: the params used to do the image
            transformation. Will be used as the key for the cache dict
        :param pixbuf: the pixbuf to be cached.
        :type pixbuf: :class:`GdkPixbuf.Pixbuf`
        """
        self._cache[params] = pixbuf
        self._waiting.discard(params)

        # Free anything that is not needed anymore from the memory
        for params in set(self._cache) - set(self._mru):
            del self._cache[params]

    def _transform_task(self):
        """Task responsible for doing image transformations.

        This will run on another thread, checking the queue for any
        new images, transforming and caching them after.

        After loading any image here, 'image-loaded' signal
        will be emitted.
        """
        while True:
            params = self._queue.get()
            # It probably isn't needed anymore
            if params not in self._mru:
                continue

            pixbuf = self._transform_image(*params)
            if pixbuf is None:
                continue

            with self._lock:
                self._cache_pixbuf(params, pixbuf)
            GObject.idle_add(self.emit, 'image-loaded')

    def _transform_image(self, path, size, fill_image, draw_border, draft):
        """Render path into a pixbuf.

        :param str path: the image path or `None` to use a fallback image
        :param int size: the size to resize the image. It will be resized
            to fit a square of (size, size)
        :param bool fill_image: if we should fill the image with a transparent
            background to make a smaller image be at least a square of
            (size, size), with the real image at the center.
        :param bool draw_border: if we should add a border on the image
        :param bool draft: if we should load the image as a draft. This
            trades a little quality for a much higher performance.
        :returns: the resized pixbuf
        :rtype: :class:`GdkPixbuf.Pixbuf`
        """
        path = path or ''
        image = self._open_image(path, size, draft)
        if image is None:
            return None

        if draw_border:
            image = add_border(image, border_size=self.IMAGE_BORDER_SIZE)
            image = add_drop_shadow(
                image, border_size=self.IMAGE_SHADOW_SIZE,
                offset=(self.IMAGE_SHADOW_OFFSET, self.IMAGE_SHADOW_OFFSET))

            size += self.IMAGE_BORDER_SIZE * 2
            size += self.IMAGE_SHADOW_SIZE * 2
            size += self.IMAGE_SHADOW_OFFSET
        else:
            # FIXME: There's a bug on PIL where image.thumbnail modifications
            # will be lost for some images when saving it the way we do on
            # image2pixbuf (even image.copy().size != image.size when it was
            # resized).  Adding a border of size 0 will make it at least be
            # pasted to a new image (which didn't have its thumbnail method
            # called), working around this issue.
            image = add_border(image, 0)

        pixbuf = image2pixbuf(image)
        width = pixbuf.get_width()
        height = pixbuf.get_height()

        if not fill_image:
            return pixbuf

        # Make sure the image is on the center of the image_max_size
        square_pic = GdkPixbuf.Pixbuf.new(
            GdkPixbuf.Colorspace.RGB, True, pixbuf.get_bits_per_sample(),
            size, size)
        # Fill with transparent white
        square_pic.fill(0xffffff00)

        dest_x = (size - width) / 2
        dest_y = (size - height) / 2
        pixbuf.copy_area(0, 0, width, height, square_pic, dest_x, dest_y)

        return square_pic

    def _open_image(self, path, size, draft):
        """Open the image on the given path.

        :param str path: the image path
        :param int size: the size to resize the image. It will be resized
            to fit a square of (size, size)
        :param bool draft: if we should load the image as a draft. This
            trades a little quality for a much higher performance.
        :returns: the opened image
        :rtype: :class:`PIL.Image`
        """
        # When trying to open the brokensuit images
        # (https://code.google.com/p/javapng/wiki/BrokenSuite), PIL failed to
        # open 27 of them, while Pixbuf failed to open 32. But trying PIL first
        # and Pixbuf if it failed reduced that number to 20.
        # In general, most of the images (specially if they are not broken,
        # which is something more uncommon) will be opened directly by PIL.
        try:
            image = Image.open(path)
            if draft:
                image.draft('P', (size, size))
            image.load()
        except (IOError, SyntaxError, OverflowError, struct.error) as e:
            try:
                pixbuf = GdkPixbuf.Pixbuf.new_from_file(path)
            except GLib.GError:
                return None
            else:
                image = Image.fromstring(
                    "RGB", (pixbuf.get_width(), pixbuf.get_height()),
                    pixbuf.get_pixels())

        image.thumbnail((size, size), Image.BICUBIC)
        return image



"""String utilities."""


def is_printable(char):
    """Determines whether a character can be displayed directly.

    Used for testing if some content should be treated as text or binary.

    :param char: Character to be tested
    :type char: str
    :rtype: bool
    """
    char_code = ord(char)
    return (char_code >= 32) or (9 <= char_code <= 13)


def replace_non_printable(string_):
    """Replace non-printable characters on the string with a replacement.

    Use the unicode replacement character (U+FFFD), instead of the
    non-printable ones.

    :param string_: The string to replace the characters from
    :type string_: str
    :rtype: str
    """
    return ''.join(c if is_printable(c) else u"\uFFFD" for c in string_)



"""Popupcal tests."""

import datetime
import unittest

from gi.repository import Gtk
import mock

from datagrid_gtk3.ui.popupcal import DateEntry


class FakeDialog(object):

    """Object mimicing Gtk.Dialog's api."""

    def __init__(self, retval, *args):
        """Initialize the FakeDialog object."""
        self._retval = retval
        self.vbox = Gtk.VBox()

    def run(self):
        """Mimic the run method."""
        return self._retval

    def set_decorated(self, value):
        """Stub set_decorated method."""
        pass

    def add_button(self, label, response):
        """Mimic fake_button method."""
        return Gtk.Button(label)

    def destroy(self):
        """Stub destroy method."""
        pass

    def get_action_area(self):
        """Mimic get_action_area method."""
        class FakeActionArea(object):
            def reorder_child(self, btn, order):
                pass
        return FakeActionArea()


class DateEntryTest(unittest.TestCase):

    """Tests for :class:`datagrid_gtk3.ui.popupcal.DateEntry`."""

    def setUp(self):
        """Setup widget for testing."""
        self._entry = DateEntry(Gtk.Window(), DateEntry.TYPE_START)

    def test_set_date(self):
        """set_date should set the text for the widget based on the format."""
        self.assertEqual(self._entry.get_property('text'), '')

        self._entry.set_date(datetime.datetime(2015, 8, 10))
        self.assertEqual(self._entry.get_property('text'), '10-Aug-2015 00:00')

        self._entry.set_date(datetime.datetime(2015, 8, 12, 15, 20))
        self.assertEqual(self._entry.get_property('text'), '12-Aug-2015 15:20')

        self._entry.set_date(None)
        self.assertEqual(self._entry.get_property('text'), '')

    def test_get_date(self):
        """get_date should return a datetime based on the text on the entry."""
        self._entry.set_property('text', '')
        self.assertIsNone(self._entry.get_date())

        self._entry.set_property('text', '10-Aug-2015 00:00')
        self.assertEqual(self._entry.get_date(),
                         datetime.datetime(2015, 8, 10))

        self._entry.set_property('text', '12-Aug-2015 15:20')
        self.assertEqual(self._entry.get_date(),
                         datetime.datetime(2015, 8, 12, 15, 20))

    def test_set_text(self):
        """set_text should set the text for the widget based on the format."""
        self.assertEqual(self._entry.get_property('text'), '')

        self._entry.set_text('2-Jun-2013 06:48')
        self.assertEqual(self._entry.get_date(),
                         datetime.datetime(2013, 6, 2, 6, 48))
        self.assertEqual(self._entry.get_property('text'), '2-Jun-2013 06:48')

        self._entry.set_text('02/10/2015')
        self.assertEqual(self._entry.get_date(),
                         datetime.datetime(2015, 2, 10))
        self.assertEqual(self._entry.get_property('text'), '10-Feb-2015 00:00')

    def test_get_text(self):
        """get_text should return a formated date based on the format."""
        self.assertEqual(self._entry.get_property('text'), '')

        self._entry.set_date(datetime.datetime(2015, 8, 10))
        self.assertEqual(self._entry.get_text(), '10-Aug-2015 00:00')

        self._entry.set_date(datetime.datetime(2015, 8, 12, 15, 20))
        self.assertEqual(self._entry.get_text(), '12-Aug-2015 15:20')

    def test_popup_response_ok(self):
        """DatePicker should change the date on entry on response ok."""
        self._entry.set_date(datetime.datetime(2015, 1, 2, 10, 20))

        with mock.patch('datagrid_gtk3.ui.popupcal.Gtk.Dialog') as dialog:
            fake_dialog = FakeDialog(Gtk.ResponseType.OK)
            dialog.return_value = fake_dialog

            with mock.patch.object(self._entry, 'get_date') as get_date:
                # Entry will use the actual date to populate the date picker
                get_date.return_value = datetime.datetime(2010, 2, 3)
                self._entry._popup_picker()
            self.assertEqual(self._entry.get_date(),
                             datetime.datetime(2010, 2, 3))

    def test_popup_response_clear(self):
        """DatePicker should clear the date on entry on response clear."""
        self._entry.set_date(datetime.datetime(2015, 1, 2, 10, 20))

        with mock.patch('datagrid_gtk3.ui.popupcal.Gtk.Dialog') as dialog:
            fake_dialog = FakeDialog(99)
            dialog.return_value = fake_dialog

            self.assertEqual(self._entry.get_date(),
                             datetime.datetime(2015, 1, 2, 10, 20))
            self._entry._popup_picker()
            self.assertIsNone(self._entry.get_date())



from contextlib import closing
import sqlite3
import tempfile

from datagrid_gtk3.utils.imageutils import get_icon_filename


TEST_DATA = {
    'people': {
        'metadata': [
            ('__id', 'INTEGER PRIMARY KEY'),
            ('first_name', 'TEXT'),
            ('last_name', 'TEXT'),
            ('age', 'INTEGER'),
            ('start_date', 'INTEGER'),
            ('image_path', 'TEXT')
        ],
        'data': [
            (1, 'Dee', 'Timberlake', 30, 1286755200,
             'file://' + get_icon_filename(['image'], 48)),
            (2, 'Steve', 'Austin', 35, 1318291200,
             'file://' + get_icon_filename(['calendar'], 48)),
            (3, 'Oscar', 'Goldman', 50, 1349913600, None),
            (4, 'Monica', 'Goldman', 40, 1344953660, None),
        ]
    },
    'files': {
        'metadata': [
            ('__id', 'TEXT PRIMARY KEY'),
            ('__parent', 'TEXT'),
            ('filename', 'TEXT'),
            ('flatname', 'TEXT'),
            ('children_len', 'INTEGER'),
        ],
        'data': [
            # The structure here is:
            #   file-0
            #   file-1
            #   folder-0/
            #       file-0-0
            #       file-0-1
            #   folder-1/
            #       file-1-0
            #       file-1-1
            #       folder-1-0
            #           file-1-0-0
            ('file-0', None,
             'file-0', 'file-0', 0),
            ('file-1', None,
             'file-1', 'file-1', 0),
            ('folder-0', None,
             'folder-0', 'folder-0', 2),
            ('folder-1', None,
             'folder-1', 'folder-1', 3),
            ('file-0-0', 'folder-0',
             'file-0-0', 'folder-0/file-0-0', 0),
            ('file-0-1', 'folder-0',
             'file-0-1', 'folder-0/file-0-1', 0),
            ('file-1-0', 'folder-1',
             'file-1-0', 'folder-1/file-1-0', 0),
            ('file-1-1', 'folder-1',
             'file-1-1', 'folder-1/file-1-1', 0),
            ('folder-1-0', 'folder-1',
             'folder-1-0', 'folder-1/folder-1-0', 1),
            ('file-1-0-0', 'folder-1-0',
             'file-1-0-0', 'folder-1-0/file-1-0-0', 0),
        ]
    },
}


def create_db(database):
    """Create a test SQLite DB for use with data grid tests.

    :param str database: one of :mod:`.TEST_DATA` keys
    :return: path to created DB temporary file
    :rtype: str
    """
    test_data = TEST_DATA[database]

    with tempfile.NamedTemporaryFile(delete=False) as fi:
        with closing(sqlite3.connect(fi.name)) as conn:
            with closing(conn.cursor()) as cursor:
                columns = [' '.join(md) for md in test_data['metadata']]
                cursor.execute("CREATE TABLE IF NOT EXISTS %s (%s)" % (
                    database, ', '.join(columns)))

                columns_names = [md[0] for md in test_data['metadata']]
                for data in test_data['data']:
                    insert_sql = "INSERT INTO %s (%s) VALUES (%s)" % (
                        database, ', '.join(columns_names),
                        (', '.join('?' for i in columns_names)))
                    cursor.execute(insert_sql, data)
                conn.commit()
    return fi.name



"""Datagrid utilies package."""

import logging
import sys

from gi.repository import Gtk, Gdk


def setup_logging_to_stdout():
    """Sets up logging to std out."""
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)

    root = logging.getLogger()
    root.setLevel(logging.DEBUG)
    root.addHandler(handler)


def setup_gtk_show_rules_hint(base_color='base_color'):
    """Modify gtk theme to display rules hinting.

    Different from gtk2, gtk3 rules hinting is only displayed if the
    theme implements it. This is a way to override that.

    :param base_color: the theme base color, if different from the default
    """
    style_provider = Gtk.CssProvider()
    style_provider.load_from_data("""
        GtkTreeView row:nth-child(even) {
            background-color: shade(@%s, 1.0);
        }
        GtkTreeView row:nth-child(odd) {
            background-color: shade(@%s, 0.95);
        }
    """ % (base_color, base_color))
    Gtk.StyleContext.add_provider_for_screen(
        Gdk.Screen.get_default(),
        style_provider,
        Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)



# -*- coding: utf-8 -*-

"""String utilities test cases."""

import string
import unittest

from datagrid_gtk3.utils.stringutils import (
    is_printable,
    replace_non_printable,
)


class StringUtilsTest(unittest.TestCase):

    """Tests for :mod:`datagrid.utils.stringutils`."""

    def test_printable(self):
        """Tests for :func:`datagrid.utils.stringutils.is_printable`."""
        for char in ['a', 'Z', '?', '\n']:
            self.assertTrue(is_printable(char))

        for char in string.whitespace:
            self.assertTrue(is_printable(char))

        # UTF-8 encoded non-ascii characters should be considered printable
        for char in u'žćčđš'.encode('utf-8'):
            self.assertTrue(is_printable(char))

        for char in [chr(0), chr(20), chr(30)]:
            self.assertFalse(is_printable(char))

    def test_replace_non_printable(self):
        """Test :func:`datagrid.utils.stringutils.replace_non_printable`."""
        self.assertEqual(
            replace_non_printable(
                "Some string\nWith\tsome %s non-printable, %s chars" % (
                    chr(20), chr(30))),
            u"Some string\nWith\tsome � non-printable, � chars")

        self.assertEqual(
            replace_non_printable(
                u"Ração %s para %s búfalos" % (chr(20), chr(30))),
            u"Ração � para � búfalos")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Test for external libs."""

import base64
import os
import tempfile
import unittest

from PIL import Image

from datagrid_gtk3.utils import imageutils

_MEDIA_FILES = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), os.pardir, "data", "media")
_TEST_IMAGE_BASE64 = """
AAABAAIAEBACAAEAAQCwAAAAJgAAACAgEAABAAQA6AIAANYAAAAoAAAAEAAAACAAAAABAAEAAAAA
AEAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAP///wAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAABjDAAA
UpAAAGMQAABSkAAAYwwAAAAAAAAAAAAAAAAAAAAAAAAAAAAA//8AAP//AAD//wAA//8AAP//AAAA
AQAAAAEAAAABAAAAAQAAAAEAAAABAAAAAQAA//8AAP//AAD//wAA//8AACgAAAAgAAAAQAAAAAEA
BAAAAAAAAAIAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACAAACAAAAAgIAAgAAAAIAAgACAgAAAgICA
AMDAwAAAAP8AAP8AAAD//wD/AAAA/wD/AP//AAD///8AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAA//gAAAD/+AAAAHj/gAAAAPAHgAAA8AeAAAeHAHAAAADwB4AAAPAHgAAI
cAAAAAAA+IgAAAD4iAAAD3AAAAAAAPAPAAAA8A8AAAiAAAAAAAD3iAAAAPeIAAAAiHeAAAAAiHAA
AACIcAAAAAeIcAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAD/////////////////////////////////
///////////////////////////////AEAQBwBAEAcAQBAHAEAQBwBAEAcAQBAHAEAQBwBAEAcAQ
BAH//////////////////////////////////////////////////////////w==
"""


class PILTest(unittest.TestCase):

    """Test for some PIL bugs."""

    def test_thumbnail_no_bug(self):
        """Test :meth:`PIL.thumbnail` without bugs."""
        image = Image.open(
            os.path.join(_MEDIA_FILES, 'icons', 'image.png'))
        image.load()
        self.assertEqual(image.size, (32, 32))

        image.thumbnail((10, 10), Image.BICUBIC)

        # This is a comparison with the test bellow in a file that doesn't fail
        image_copy = image.copy()
        self.assertEqual(image_copy.size, (10, 10))
        pixbuf = imageutils.image2pixbuf(image)
        self.assertEqual(
            (pixbuf.get_width(), pixbuf.get_height()), (10, 10))

    def test_thumbnail_bug_exists(self):
        """Test for the existence of a bug on :meth:`PIL.thumbnail`."""
        with tempfile.NamedTemporaryFile() as f:
            # This image has something that breaks thumbnail propagation
            f.write(base64.b64decode(_TEST_IMAGE_BASE64))
            f.flush()

            image = Image.open(f.name)
            image.load()
            self.assertEqual(image.size, (32, 32))

            image.thumbnail((10, 10), Image.BICUBIC)

            # When both those tests fail, it means we can remove fix the FIXME
            # code on datagrid_gtk3.utils.transformations.image_transform.
            # This should propagate the changes done by thumbnail
            image_copy = image.copy()
            self.assertEqual(image_copy.size, (32, 32))
            pixbuf = imageutils.image2pixbuf(image)
            self.assertEqual(
                (pixbuf.get_width(), pixbuf.get_height()), (32, 32))



if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

"""Date utilities test cases."""

import datetime
import unittest

from datagrid_gtk3.utils.dateutils import (
    InvalidDateFormat,
    parse_string,
)


class DateUtilsTest(unittest.TestCase):

    """Tests for :mod:`datagrid.utils.dateutils`."""

    def test_parse_string_valid(self):
        """Parse string should generate datetime for valid inputs."""
        for date_str, date in [
                ('4/10/2015', datetime.datetime(2015, 4, 10)),
                ('4/10/2015 04:20', datetime.datetime(2015, 4, 10, 4, 20)),
                ('2-Jun-2013', datetime.datetime(2013, 6, 2)),
                ('2-Jun-2013 06:48:15',
                 datetime.datetime(2013, 6, 2, 6, 48, 15)),
                ('Tue, 10 Apr 2001 15:51:24',
                 datetime.datetime(2001, 4, 10, 15, 51, 24))]:
            self.assertEqual(parse_string(date_str), date)

    def test_parse_string_invalid(self):
        """InvalidDateFormat should be raised for invalid inputs."""
        for invalid_str in [
                'non-valid-string',
                '10/50/2010',
                '50/10/2015',
                '10/10/2010 25:10']:
            with self.assertRaises(InvalidDateFormat):
                parse_string(invalid_str)



"""User interface file handling classes."""
from collections import defaultdict

from gi.repository import Gtk


class UIFile(object):

    """Load user interface file and provide additonal features.

    :param ui_filename: Name pointing to the user interface file
    :type ui_filename: str

    Additional features provided are as follows:

    * Connect callbacks methods from ui file
    * Attribute access to widgets in the ui file through
      ``Gtk.Builder.get_object()``
    * Helper methods to connect/disconnect all signal handlers to
      other objects. This is useful for dialogs that are destroyed,
      but connected to signals of other objects that remain alive. This
      is important because otherwise a reference to the callback
      methods is kept and, in fact, those callbacks are executed even
      after the dialog has been destroyed.
    """

    def __init__(self, ui_filename):
        """Open user interface file and connect widgets to callback methods."""
        builder = Gtk.Builder()
        builder.add_from_file(ui_filename)
        builder.connect_signals(self)
        setattr(self, 'builder', builder)

        # Keep a record of all handlers to disconnect them later if needed
        self._handler_ids = defaultdict(list)

    def __getattr__(self, name):
        """Look for widgets in builder object as attributes.

        :param name: Attribute name
        :type name: str
        :returns: Widget
        :rtype: ``Gtk.Widget`` (any of its subclasses)
        :raises:
            ``AtributeError`` when the attribute definition is not found in the
            ui file
        """
        obj = self.builder.get_object(name)

        if obj is None:
            raise AttributeError(name)

        return obj

    def connect_signal(self, obj, signal, handler, *args):
        """Connect signal handler and keep a record of its id.

        The id of the handlers is used to disconnect it on destroy

        :param obj: Object that will emit the signal
        :type obj: ``GObject.GObject``
        :param signal: Signal name
        :type signal: str
        :param handler: Callback to be executed when the signal is emitted
        :type handler: callable
        :param args: Additional arguments required by the signal
        :type args: iterable
        """
        handler_id = obj.connect(signal, handler, *args)
        self._handler_ids[obj].append(handler_id)
        return handler_id

    def disconnect_all_signals(self):
        """Disconnect all handlers."""
        for obj, handler_ids in self._handler_ids.iteritems():
            for handler_id in handler_ids:
                if obj.handler_is_connected(handler_id):
                    obj.disconnect(handler_id)


class SignalBlocker:

    """Block widget signals connected to a callback.

    :param widget: Widget that will emit the signal
    :type widget: ``GObject.GObject``
    :param callback: Function to block
    :type callback: callable

    """

    def __init__(self, widget, callback):
        """Keep a reference to widget and callback.

        Those references are required later when using the signal blocker as a
        context manager.

        """
        self.widget = widget
        self.callback = callback

    def __enter__(self):
        """Block sighnal handler for a given widget."""
        self.widget.handler_block_by_func(self.callback)

    def __exit__(self, type, value, traceback):
        """Unblock sighnal handler for a given widget."""
        self.widget.handler_unblock_by_func(self.callback)
        return False



import contextlib
import os
import sqlite3
import unittest

from datagrid_gtk3.tests.data import create_db, TEST_DATA
from datagrid_gtk3.db.sqlite import SQLiteDataSource


class SQLiteDataSourceTest(unittest.TestCase):

    """Test SQLiteDataSource."""

    def setUp(self):
        """Create test data."""
        self.table = 'people'
        self.db_file = create_db(self.table)
        self.datasource = SQLiteDataSource(
            self.db_file,
            table=self.table,
            config=[
                {'column': 'First name', 'type': 'str'},
                {'column': 'Last name', 'type': 'str'},
                {'column': 'Age', 'type': 'int'},
                {'column': 'Start', 'type': 'int', 'encoding': 'timestamp'},
                {'column': 'Image', 'type': 'int', 'encoding': 'image'},
            ],
        )
        self.datasource.MAX_RECS = 2

    def tearDown(self):
        """Remove test data file."""
        os.unlink(self.db_file)

    def test_load(self):
        """Load first page of records and get total."""
        rows = self.datasource.load()
        self.assertEqual(len(rows), 2)
        self.assertEqual(self.datasource.total_recs, 4)

    def test_load_with_params(self):
        """Filter and order records."""
        param = {
            'where': {
                'age': {
                    'param': 30, 'operator': '>'
                }
            },
            'order_by': 'age',
            'desc': True
        }
        rows = self.datasource.load(param)
        self.assertEqual(len(rows), 2)
        self.assertEqual(rows[0].data[2], 'Goldman')

    def test_load_with_where_param(self):
        """Filter results with a search param."""
        param = {
            'where': {
                'search': {
                    'param': 'gold',
                },
            },
        }
        rows = self.datasource.load(param)
        self.assertEqual(len(rows), 2)
        self.assertEqual(
            {('Oscar', 'Goldman'), ('Monica', 'Goldman')},
            {(row.data[1], row.data[2]) for row in rows})

    def test_load_paging(self):
        """Load first and second pages of records."""
        self.datasource.load()  # initial load is always without paging
        rows = self.datasource.load({'page': 1})
        self.assertEqual(len(rows), 2)
        self.assertEqual(rows[0].data[2], 'Goldman')

    def test_update(self):
        """Update __selected in first record in data set."""
        self.datasource.update({'__selected': True}, [1])
        # ^^ update row with id 1
        rows = self.datasource.load()
        self.assertEqual(rows[0].data[0], 1)

    def test_get_all_record_ids(self):
        """Get all record ids for a particular query."""
        param = {
            'where': {
                'age': {
                    'param': 30, 'operator': '>'
                }
            }
        }
        ids = self.datasource.get_all_record_ids(param)
        self.assertEqual(ids, [2, 3, 4])

    def test_visible_columns(self):
        """Set visible columns and ensure they're persisted."""
        self.datasource.set_visible_columns(['last_name'])
        self.assertEqual(self.datasource.get_visible_columns(), ['last_name'])

    def test_get_single_record(self):
        """Retrieve a single record as a tuple of values."""
        row = self.datasource.get_single_record(1)
        self.assertEqual(row[0], 1)
        self.assertEqual(row[1], 'Dee')
        self.assertEqual(row[2], 'Timberlake')

    def test_select(self):
        """Get data without class instance or paging."""
        db_file = self.datasource.db_file
        with contextlib.closing(sqlite3.connect(db_file)) as conn:
            results = list(self.datasource.select(conn, self.datasource.table))
        self.assertEqual(len(results), 4)

    def test_explicit_query(self):
        """Test using an explicit query for the data source."""
        # Important to not ensure "selected" column if there is no primary
        # key returned by the query.
        datasource = SQLiteDataSource(
            self.db_file, query='SELECT first_name, age FROM people',
            ensure_selected_column=False
        )
        rows = datasource.load()
        data = {tuple(row.data) for row in rows}
        reference_data = {(record[1], record[3])
                          for record in TEST_DATA[self.table]['data']}
        self.assertEqual(data, reference_data)


if __name__ == '__main__':
    unittest.main()



"""Data grid MVC package.

Usage::

    from gi.repository import Gtk
    from datagrid_gtk3.ui.grid import DataGridContainer, DataGridController
    from datagrid_gtk3.db.sqlite import SQLiteDataSource
    win = Gtk.Window()
    datagrid_container = DataGridContainer(win)
    datagrid_source = SQLiteDataSource(
        '/path/to/sqlite.db',
        'table_name',
        # optional table config list arg
    )
    datagrid_controller = DataGridController(datagrid_container, datagrid_source)
    win.add(datagrid_container.vpaned_grid)
    win.show_all()

"""



"""Data transformation utils."""

import datetime
import logging
import HTMLParser

from decimal import Decimal

import dateutil.parser
from gi.repository import Gtk

from datagrid_gtk3.utils import imageutils
from datagrid_gtk3.utils import dateutils
from datagrid_gtk3.utils import stringutils

logger = logging.getLogger(__name__)
_transformers = {}

__all__ = ('get_transformer', 'register_transformer')


def get_transformer(transformer_name):
    """Get transformation for the given name.

    :param str transformer_name: the name of the registered transformer
    :return: the transformer registered by transformer_name
    :rtype: callable
    """
    return _transformers.get(transformer_name, None)


def register_transformer(transformer_name, transformer):
    """Register a transformer.

    :param str transformer_name: the name to register the transformer
    :param callable transformer: the transformer to be registered
    """
    assert callable(transformer)
    _transformers[transformer_name] = transformer


def unregister_transformer(transformer_name):
    """Unregister a transformer.

    :param str transformer_name: the name to register the transformer
    :raise KeyError: if a transformer is not registered under the given name
    """
    del _transformers[transformer_name]


def transformer(transformer_name):
    """A decorator to easily register a decorator.

    Use this like::

        @transformer('transformer_name')
        def transformer_func(value):
            return do_something_with_value()

    :param str transformer_name: the name to register the transformer
    """
    def _wrapper(f):
        register_transformer(transformer_name, f)
        return f
    return _wrapper


###
# Default transformers
###


@transformer('string')
def string_transform(value, max_length=None, oneline=True,
                     decode_fallback=None):
    """String transformation.

    :param object value: the value that will be converted to
        a string
    :param int max_length: if not `None`, will be used to
        ellipsize the string if greater than that.
    :param bool oneline: if we should join all the lines together
        in one line
    :param callable decode_fallback: a callable to use
        to decode value in case it cannot be converted to unicode directly
    :return: the string representation of the value
    :rtype: str
    """
    if value is None:
        return '<NULL>'

    if isinstance(value, str):
        value = unicode(value, 'utf-8', 'replace')
    else:
        try:
            value = unicode(value)
        except UnicodeDecodeError:
            if decode_fallback is None:
                raise
            value = decode_fallback(value)

    # Replace non-printable characters on the string so the user will
    # know that there's something there even though it is not printable.
    value = stringutils.replace_non_printable(value)

    if oneline:
        value = u' '.join(v.strip() for v in value.splitlines() if v.strip())

    # Don't show more than max_length chars in treeview. Helps with performance
    if max_length is not None and len(value) > max_length:
        value = u'%s [...]' % (value[:max_length], )

    # At the end, if value is unicode, it needs to be converted to
    # an utf-8 encoded str or it won't be rendered in the treeview.
    return value.encode('utf-8')


@transformer('html')
def html_transform(value, max_length=None, oneline=True,
                   decode_fallback=None):
    """HTML transformation.

    :param object value: the escaped html that will be unescaped
    :param int max_length: if not `None`, will be used to
        ellipsize the string if greater than that.
    :param bool oneline: if we should join all the lines together
        in one line
    :param callable decode_fallback: a callable to use
        to decode value in case it cannot be converted to unicode directly
    :return: the html string unescaped
    :rtype: str
    """
    if value is None:
        return '<NULL>'

    html_parser = HTMLParser.HTMLParser()
    unescaped = html_parser.unescape(value)
    return string_transform(
        unescaped, max_length=max_length, oneline=oneline,
        decode_fallback=decode_fallback)


@transformer('boolean')
def boolean_transform(value):
    """Transform boolean values to a gtk stock image.

    :param bool value: the value to transform
    :return: a pixbuf representing the value's bool value
    :rtype: :class:`GdkPixbuf.Pixbuf`
    """
    img = Gtk.Image()
    # NOTE: should be STOCK_NO instead of STOCK_CANCEL but it looks
    # crappy in Lubuntu
    return img.render_icon(
        Gtk.STOCK_YES if value else Gtk.STOCK_CANCEL, Gtk.IconSize.MENU)


@transformer('bytes')
def bytes_transform(value):
    """Transform bytes into a human-readable value.

    :param int value: bytes to be humanized
    :returns: the humanized bytes
    :rtype: str
    """
    if value is None:
        return ''

    for suffix, factor in [
            ('PB', 1 << 50),
            ('TB', 1 << 40),
            ('GB', 1 << 30),
            ('MB', 1 << 20),
            ('kB', 1 << 10),
            ('B', 0)]:
        if value >= factor:
            value = '%.*f %s' % (1, float(value) / max(factor, 1), suffix)
            break
    else:
        raise ValueError('Unexpected value: %s' % (value, ))

    return value


@transformer('datetime')
def datetime_transform(value):
    """Transform datetime to ISO 8601 date format.

    :param value: the datatime object
    :type value: datetime.datetime
    :return: the datetime represented in ISO 8601 format
    :rtype: str
    """
    if value is None:
        return ''

    if isinstance(value, basestring):
        try:
            # Try to parse string as a date
            value = dateutil.parser.parse(value)
        except (OverflowError, TypeError, ValueError):
            pass

    # FIXME: Fix all places using 'datetime' for timestamp
    # (either as an int/long or as a convertable str)
    try:
        long_value = long(value)
    except (TypeError, ValueError):
        pass
    else:
        return timestamp_transform(long_value)

    if not isinstance(value, datetime.datetime):
        # Convert value to string even if it cannot be parsed as a datetime
        logger.warning('Not a datetime: %s', value)
        return str(value)

    return value.isoformat(' ')


@transformer('timestamp')
@transformer('timestamp_unix')
def timestamp_transform(value, date_only=False):
    """Transform timestamp to ISO 8601 date format.

    :param int value: Unix timestamp
    :param bool date_only: if we should format only the date part,
         ignoring the time
    :return: the datetime represented in ISO 8601 format
    :rtype: str
    """
    if value is None:
        return ''

    try:
        dt = datetime.datetime.utcfromtimestamp(value)
    except (TypeError, ValueError):
        # Convert value to string even if it cannot be parsed as a timestamp
        logger.warning('Not a timestamp: %s', value)
        return str(value)

    if date_only:
        return dt.date().isoformat()
    else:
        return dt.isoformat(' ')


@transformer('timestamp_ms')
@transformer('timestamp_unix_ms')
def timestamp_ms_transform(value):
    """Transform timestamp in milliseconds to ISO 8601 date format.

    :param int value: Unix timestamp in milliseconds
    :return: the datetime represented in ISO 8601 format
    :rtype: str
    """
    if value is None:
        return ''

    return timestamp_transform(
        dateutils.normalize_timestamp(value, 'timestamp_unix_ms'))


@transformer('timestamp_Ms')
@transformer('timestamp_unix_Ms')
def timestamp_Ms_transform(value):
    """Transform timestamp in microseconds to ISO 8601 date format.

    :param int value: Unix timestamp in microseconds
    :return: the datetime represented in ISO 8601 format
    :rtype: str
    """
    if value is None:
        return ''

    return timestamp_transform(
        dateutils.normalize_timestamp(value, 'timestamp_unix_Ms'))


@transformer('timestamp_ios')
@transformer('timestamp_apple')
def timestamp_apple_transform(value):
    """Transform apple timestamp to ISO 8601 date format.

    Apple timestamps (e.g. those used on iOS) start at 2001-01-01.

    :param int value: apple timestamp
    :return: the datetime represented in ISO 8601 format
    :rtype: str
    """
    if value is None:
        return ''

    return timestamp_transform(
        dateutils.normalize_timestamp(value, 'timestamp_apple'))


@transformer('timestamp_webkit')
def timestamp_webkit_transform(value):
    """Transform WebKit timestamp to ISO 8601 date format.

    WebKit timestamps are expressed in microseconds and
    start at 1601-01-01.

    :param int value: WebKit timestamp
    :return: the datetime represented in ISO 8601 format
    :rtype: str
    """
    if value is None:
        return ''

    return timestamp_transform(
        dateutils.normalize_timestamp(value, 'timestamp_webkit'))


@transformer('timestamp_julian')
def timestamp_julian_transform(value, date_only=False):
    """Transform Julian timestamp to ISO 8601 date format.

    Julian timestamps are the number of days that has passed since
    noon Universal Time on January 1, 4713 BCE.

    :param int value: Julian timestamp in days
    :param bool date_only: if we should format only the date part,
         ignoring the time
    :return: the datetime represented in ISO 8601 format
    :rtype: str
    """
    if value is None:
        return ''

    return timestamp_transform(
        dateutils.normalize_timestamp(value, 'timestamp_julian'),
        date_only=date_only)


@transformer('timestamp_julian_date')
def timestamp_julian_date_transform(value):
    """Transform julian timestamp to ISO 8601 date format.

    Julian timestamps are the number of days that has passed since
    noon Universal Time on January 1, 4713 BCE.

    :param int value: Julian timestamp
    :return: the date represented in ISO 8601 format
    :rtype: str
    """
    if value is None:
        return ''

    return timestamp_julian_transform(value, date_only=True)


@transformer('timestamp_midnight')
def timestamp_midnight_transform(value):
    """Transform midnight timestamp to ISO 8601 time format.

    Midnight timestamp is the count in seconds of the time that
    has passed since midnight.

    :param int value: midnight timestamp in seconds
    :return: the time represented in ISO 8601 format
    :rtype: str
    """
    if value is None:
        return ''

    dt = datetime.datetime.min + datetime.timedelta(0, value)
    return dt.time().isoformat()


@transformer('timestamp_midnight_ms')
def timestamp_midnight_ms_transform(value):
    """Transform midnight timestamp in milliseconds to ISO 8601 time format.

    Midnight timestamp is the count in seconds of the time that
    has passed since midnight.

    :param int value: midnight timestamp in milliseconds
    :return: the time represented in ISO 8601 format
    :rtype: str
    """
    if value is None:
        return ''

    return timestamp_midnight_transform(value / 10 ** 3)


@transformer('timestamp_midnight_Ms')
def timestamp_midnight_Ms_transform(value):
    """Transform midnight timestamp in microsecond to ISO 8601 time format.

    Midnight timestamp is the count in seconds of the time that
    has passed since midnight.

    :param int value: midnight timestamp in microseconds
    :return: the time represented in ISO 8601 format
    :rtype: str
    """
    if value is None:
        return ''

    return timestamp_midnight_transform(value / 10 ** 6)


@transformer('image')
def image_transform(path, size=24, fill_image=True, draw_border=False,
                    draft=False, load_on_thread=False):
    """Render path into a pixbuf.

    :param str path: the image path or `None` to use a fallback image
    :param int size: the size to resize the image. It will be resized
        to fit a square of (size, size)
    :param bool fill_image: if we should fill the image with a transparent
        background to make a smaller image be at least a square of
        (size, size), with the real image at the center.
    :param bool draw_border: if we should add a border on the image
    :param bool draft: if we should load the image as a draft. This
        trades a little quality for a much higher performance.
    :param bool load_on_thread: if we should load the image on another
        thread. This will make a placeholder be returned the first
        time this method is called.
    :returns: the resized pixbuf
    :rtype: :class:`GdkPixbuf.Pixbuf`
    """
    cm = imageutils.ImageCacheManager.get_default()
    return cm.get_image(path, size, fill_image, draw_border,
                        draft, load_on_thread)


@transformer('degree_decimal_str')
def degree_decimal_str_transform(value, length=8):
    """Transform degree decimal string to a numeric value.

    The string is expected to have <length> digits, if less digits are found,
    it will be prefixed with zeroes as needed.

    :param value: Degrees encoded as a string with digits
    :type value: str
    :param length: Maximum expected string length
    :type length: int

    """
    assert isinstance(value, basestring), 'String value expected'
    assert value.isdigit(), 'All characters expected to be digits'
    assert len(value) <= length, \
        'String length expected to be {} or less'.format(length)
    value = value.zfill(length)

    # Add decimal point at the expected location
    value = '{}.{}'.format(value[:2], value[2:])

    # Remove non-significant leading zeroes
    value = Decimal(value)
    return str(value)



"""SQLite database backend."""

import logging
import operator
import sqlite3
import struct
import weakref
from contextlib import closing

from gi.repository import GObject
from sqlalchemy import (
    Column,
    INTEGER,
    MetaData,
    Table,
    create_engine,
    inspect,
)
from sqlalchemy.exc import DatabaseError
from sqlalchemy.sql import (
    cast,
    alias,
    and_,
    collate,
    column,
    desc,
    func,
    or_,
    select,
    table as table_,
)

from datagrid_gtk3.db import DataSource, Node

logger = logging.getLogger(__name__)
_compile = lambda q: q.compile(compile_kwargs={"literal_binds": True}).string

_OPERATOR_MAPPER = {
    'is': operator.eq,
    '=': operator.eq,
    '!=': operator.ne,
    '<': operator.lt,
    '<=': operator.le,
    '>': operator.gt,
    '>=': operator.ge,
}


class SQLiteDataSource(DataSource):

    """SQLite data source especially for use with a `Gtk.TreeModel`.

    Provides a SQLite backend for providing data to a
    :class:`datagrid_gtk3.ui.grid.DataGridModel` instance, which is
    a GTK `TreeModel`.

    Optional table configuration example::

        [
            ('ID', (int, None)),
            ('Title', (str, None)),
            ('Date', (long, 'datetime')),
            ('Thumbnail', (buffer, 'image'))
        ]

    :param str db_file: path to SQLite database file
    :param str table: name of table in SQLite db
    :param str update_table: table to perform update operations on, eg.
        if the table being SELECTed is actually a view
    :param list config: list of table configuration tuples including display
        names, data types, transforms, etc.
    :param bool ensure_selected_column: Whether to ensure the presence of
        the __selected column.
    :param bool display_all: Whether or not all columns should be displayed.
    :param str query: Full custom query to be used instead of the table name.
    :param bool persist_columns_visibility: Weather we should persist
        the columns visibility in the database.
    """

    __gsignals__ = {
        'rows-changed': (GObject.SignalFlags.RUN_LAST, None, (object, object)),
    }

    _DBS = weakref.WeakSet()

    MAX_RECS = 100
    SQLITE_PY_TYPES = {
        'INT': long,
        'INTEGER': long,
        'LONG': long,
        'TEXT': str,
        'REAL': float,
        'BLOB': str
    }

    STRING_PY_TYPES = {  # NOTE: ideally could use eval, but unsafe
        'int': int,
        'long': long,
        'str': str,
        'float': float,
        'buffer': buffer
    }

    def __init__(self, db_file, table=None, update_table=None, config=None,
                 ensure_selected_column=True,
                 display_all=False, query=None,
                 persist_columns_visibility=True):
        """Process database column info."""
        super(SQLiteDataSource, self).__init__()

        assert table or query  # either table or query must be given
        self.db_file = db_file
        self.table = table_(table if table else "__CustomQueryTempView")
        self.query = query
        if query:
            logger.debug("Custom SQL: %s", query)
        self._persist_columns_visibility = persist_columns_visibility
        self._ensure_selected_column = ensure_selected_column
        self.display_all = display_all
        # FIXME: Use sqlalchemy for queries using update_table
        if update_table is not None:
            self.update_table = update_table
        else:
            self.update_table = table
        self.config = config
        self.columns = self.get_columns()
        self.columns_idx = {
            col['name']: i for i, col in enumerate(self.columns)}

        for col in self.columns:
            self.table.append_column(column(col['name']))

        self.visible_columns_table = table_('__visible_columns')
        for col in ['tablename', 'columns']:
            self.visible_columns_table.append_column(column(col))

        with closing(sqlite3.connect(self.db_file)) as conn:
            with closing(conn.cursor()) as cursor:
                # FIXME: Maybe we should use a parameter to generate
                # search_table if it doesn't exist?
                search_table = self.table.name + '_search'
                cursor.execute('PRAGMA table_info(%s)' % (search_table, ))
                if cursor.fetchone():
                    self.search_table = search_table
                else:
                    self.search_table = None

                # Migrate old `_selected_columns` to `__visible_columns`
                cursor.execute('PRAGMA table_info(_selected_columns)')
                columns = {column_info[1] for column_info in cursor.fetchall()}
                if columns == {u'columns', u'tablename'}:
                    # `_selected_columns` exists and has the appropriate schema
                    cursor.execute('ALTER TABLE _selected_columns '
                                   'RENAME TO __visible_columns')
                    conn.commit()

        self.__class__._DBS.add(self)

    ###
    # Public
    ###

    def load(self, params=None):
        """Execute SQL ``SELECT`` and populate ``rows`` attribute.

        Loads a maximum of ``MAX_RECS`` records at a time.

        ``params`` dict example::

            {
                'desc': False,
                'order_by': 'title',
                'where': {
                    'date': {
                        'operator': 'range',
                        'param': (0, 1403845140)
                    },
                    'search': {
                        'operator': '=',
                        'param': 'Google'}
                    }
                }
            }

        :param dict params: dict of various parameters from which to construct
            additional SQL clauses eg. ``WHERE``, ``ORDER BY``, etc.
        """
        rows = Node()
        # FIXME: Maybe we should use kwargs instead of params?
        params = params or {}

        # WHERE
        where = params.get('where', None)
        if where is not None:
            where = self._get_where_clause(where)

        # ORDER BY
        order_by = params.get('order_by', None)
        # Do a numeric ordering first, as suggested here
        # (http://stackoverflow.com/a/4204641), and then a case-insensitive one
        order_by = (order_by and
                    [self.table.columns[order_by] + 0,
                     collate(self.table.columns[order_by], 'NOCASE')])
        if order_by is not None and params.get('desc', False):
            order_by = [desc(col) for col in order_by]

        # OFFSET
        page = params.get('page', 0)
        offset = page * self.MAX_RECS
        # A little optimization to avoid doing more queries when we
        # already loaded everything
        if page > 0 and offset >= self.total_recs:
            return rows

        # Flat
        flat = params.get('flat', False)
        if flat:
            flat_where = operator.ne(
                self.table.columns[self.FLAT_COLUMN], None)
            where = and_(where, flat_where) if where is not None else flat_where  # noqa

        with closing(sqlite3.connect(self.db_file)) as conn:
            conn.row_factory = lambda cursor, row: list(row)
            # ^^ make result lists mutable so we can change values in
            # the GTK TreeModel that uses this datasource.
            conn.create_function('rank', 1, rank)
            # TODO: ^^ only if search term in params
            with closing(conn.cursor()) as cursor:
                self._ensure_temp_view(cursor)

            if page == 0:
                # set the total record count the only the first time the
                # record set is requested
                res = self.select(
                    conn, self.table, [func.count(1)], where=where)
                self.total_recs = int(list(res)[0][0])

            if self.PARENT_ID_COLUMN and not flat:
                rows.extend(
                    self._load_tree_rows(
                        conn, where, order_by, params.get('parent_id', None)))
            else:
                query = self.select(
                    conn, self.table, self.table.columns, where=where,
                    limit=self.MAX_RECS, offset=offset, order_by=order_by)
                for row in query:
                    rows.append(Node(data=row))

        rows.children_len = len(rows)
        return rows

    def update(self, params, ids=None):
        """Update the recordset with a SQL ``UPDATE`` statement.

        Typically used to update the ``__selected`` column indicating
        selected records.

        If `ids` is None, will update the entire table.

        :param dict params: keys corresponding to DB columns + values to update
        :param list ids: database primary keys to use for updating
        """
        # FIXME: Use sqlalchemy to construct the queries here
        with closing(sqlite3.connect(self.db_file)) as conn:
            with closing(conn.cursor()) as cursor:
                update_sql_list = []
                for key, value in params.iteritems():
                    if isinstance(value, bool):
                        value = int(value)
                    elif isinstance(value, basestring):
                        value = "'%s'" % value
                    update_sql_list.append('%s=%s' % (key, value))
                update_sql_str = ', '.join(update_sql_list)
                if ids is not None:
                    for id_ in ids:
                        sql = 'UPDATE %s SET %s WHERE %s = ?' % (
                            self.update_table, update_sql_str, self.ID_COLUMN)
                        cursor.execute(sql, (str(id_),))
                else:
                    sql = 'UPDATE %s SET %s' % (
                        self.update_table, update_sql_str)
                    cursor.execute(sql)
                conn.commit()

        # Emit rows-changed for any other databases connected to the same
        # database and table. This is to allow any view using them
        # to update themselves with the changes done here.
        # The current object will not emit the event as it is the one who
        # made the update and thus, is assumed to know about the changes
        for db in self.__class__._DBS:
            if db is self:
                continue
            if (db.db_file, db.table.name) != (self.db_file, self.table.name):
                continue

            db.emit('rows-changed', params, ids)

    def get_all_record_ids(self, params=None):
        """Get all the record primary keys for given params.

        :param dict params: params from which to construct SQL ``WHERE`` clause
        :return: primary key ids
        :rtype: list
        """
        with closing(sqlite3.connect(self.db_file)) as conn:
            conn.create_function('rank', 1, rank)
            # TODO: ^^ create this function only if search term in params
            where = params and params.get('where', None)
            if where is not None:
                where = self._get_where_clause(where)
            res = self.select(
                conn, self.table,
                [self.table.columns[self.ID_COLUMN]], where=where)

            return [row[0] for row in res]

    def get_single_record(self, record_id):
        """Get single record from database for display in preview pane.

        :param int record_id: required record number to be retrieved
        :return: row of data
        :rtype: tuple
        """
        with closing(sqlite3.connect(self.db_file)) as conn:
            conn.row_factory = sqlite3.Row  # Access columns by name
            res = list(self.select(
                conn, self.table, self.table.columns,
                where=self.table.columns[self.ID_COLUMN] == record_id))

            # TODO log error if more than one
            return res[0]

    def get_visible_columns(self):
        """Get visible columns info from DB.

        :returns: list of column names
        :rtype: list or None
        """
        if not self._persist_columns_visibility:
            return None

        where = (self.visible_columns_table.columns['tablename'] ==
                 self.table.name)
        columns = [self.visible_columns_table.columns['columns']]

        with closing(sqlite3.connect(self.db_file)) as conn:
            conn.row_factory = sqlite3.Row  # Access columns by name
            try:
                result = list(
                    self.select(conn, self.visible_columns_table,
                                columns, where=where))
            except sqlite3.OperationalError as err:
                # FIXME: When will this happen?
                logger.warn(str(err))
                return None

        if not result:
            return None

        return result[0][0].split(',')

    def set_visible_columns(self, columns):
        """Update the *__visible_columns* table.

        Updates the table in the DB that stores info about which columns have
        been selected.  This is used to exclude unwanted columns from a report.

        :param list columns: list of column names to display
        """
        if not self._persist_columns_visibility:
            return

        with closing(sqlite3.connect(self.db_file)) as conn:
            with closing(conn.cursor()) as cursor:
                table = self.visible_columns_table.name
                cursor.execute(
                    'CREATE TABLE IF NOT EXISTS %s '
                    '(tablename TEXT PRIMARY KEY, columns TEXT)' % (table, ))
                cursor.execute(
                    'INSERT OR REPLACE INTO %s (tablename, columns) '
                    'VALUES (?, ?)' % (table, ),
                    (self.table.name, ','.join(columns)))
                conn.commit()

    def select(self, conn, table, columns=None, where=None,
               order_by=None, limit=None, offset=None):
        """Select records from given db and table given columns and criteria.

        :param str db_file: path to SQLite database file
        :param str table: name of table in SQLite db
        :param list columns: list of columns to SELECT from
        :param dict where: dict of parameters to build ``WHERE`` clause
        """
        columns = columns or table.columns
        sql = select(
            columns=columns, whereclause=where,
            from_obj=[table], order_by=order_by)
        sql_str = _compile(sql)

        # XXX: How to make sqlalchemy use limit/offset right? It is not
        # replacing the values on _compile
        if limit is not None:
            sql_str += '\nLIMIT %s' % (limit, )
        if offset is not None:
            sql_str += '\nOFFSET %s' % (offset, )

        logger.debug('SQL:\n%s', sql_str)
        with closing(conn.cursor()) as cursor:
            for row in cursor.execute(sql_str):
                yield row

    ###
    # Private
    ###

    def _get_where_clause(self, where_params):
        """Construct a SQL ``WHERE`` clause.

        A typical ``where_params`` dict might look like this::

            {'search': {'operator': '=', 'param': 'Google'}}

        .. NOTE:: ``search`` is a special key used for full-text searches

        :param dict where_params: parameters to build ``WHERE`` clause
        :return: SQL ``WHERE`` clause, and parameters to use in clause
        :rtype: tuple
        """
        sql_clauses = []
        for key, value in where_params.iteritems():
            if key == 'search':
                # full-text search
                if value['param'] and self.search_table is not None:
                    # XXX: This is to make MATCH be compiled direct here.
                    # We should build this query using sqlalchemy instead
                    match = column(self.search_table).match(value['param'])

                    sql = '(%s IN (%s)' % (
                        self.ID_COLUMN,
                        'SELECT %(id)s FROM '
                        '(SELECT rank(matchinfo(%(table)s)) AS r, %(id)s'
                        ' FROM  %(table)s WHERE %(match)s)'
                        ' WHERE r > 0 ORDER BY r DESC)' % {
                            "id": self.ID_COLUMN,
                            "table": self.search_table,
                            "match": _compile(match),
                        }
                    )
                    sql_clauses.append(sql)
                elif value['param']:
                    clauses = [col.like('%{}%'.format(value['param']))
                               for col in self.table.columns]
                    sql_clauses.append(or_(*clauses))
            elif value['operator'] == 'range':
                sql_clauses.append(
                    self.table.columns[key].between(*value['param']))
            else:
                clause = _OPERATOR_MAPPER[value['operator']](
                    self.table.columns[key], value['param'])
                sql_clauses.append(clause)

        return and_(*sql_clauses)

    def _ensure_temp_view(self, cursor):
        """If a custom query is defined, temporary view using that query
        is used in place of a table name.
        This makes sure that temporary view exists if required.

        :param cursor: Cursor for the session where the view might be needed.
        """
        if self.query:
            # create a temporary view for collecting column info
            cursor.execute('CREATE TEMP VIEW IF NOT EXISTS %s AS %s' % (
                self.table.name, self.query
            ))

    def _ensure_primary_key_column(self, conn):
        """Ensure that we know what is the primary key.

        :param conn: an open connection to the database
        """
        # FIXME: What to do when using temporary views?
        if self.query:
            return True

        with closing(conn.cursor()) as cursor:
            cursor.execute('PRAGMA table_info(%s)' % (self.table.name, ))
            rows = cursor.fetchall()

        row_names = [row[1] for row in rows]

        # First check if there's any row that matches self.ID_COLUMN
        if self.ID_COLUMN in row_names:
            return True

        # Then try to find any row that has its primary key flag set to True
        for row in rows:
            if row[5]:  # primary key
                self.ID_COLUMN = row[1]
                return True

        self.ID_COLUMN = '_rowid_'
        return False

    def get_columns(self):
        """Return a list of column information dicts.

        Queries either the database ``PRAGMA`` for column information or
        uses the config information passed into the constructor.

        Column dict example::

            {
                'transform': None,
                'type': str,
                'name': 'title',
                'display': 'Title'
            }

        :return: a list of column information dicts
        :rtype: list
        """
        cols = []
        with closing(sqlite3.connect(self.db_file)) as conn:
            has_primary_key = self._ensure_primary_key_column(conn)

            with closing(conn.cursor()) as cursor:
                self._ensure_temp_view(cursor)
                table_info_query = 'PRAGMA table_info(%s)' % self.table.name
                cursor.execute(table_info_query)
                rows = cursor.fetchall()
                # If the table doesn't have a real primary key,
                # we will use its rowid, but it is not present on table_info
                if not has_primary_key:
                    rows.append(
                        (len(rows), self.ID_COLUMN, 'INTEGER', 0, '', 1))

                skip_config = [self.ID_COLUMN, self.SELECTED_COLUMN]
                has_selected = False
                counter = 0
                for i, row in enumerate(rows):
                    col_name = row[1]
                    if self.config is not None and col_name not in skip_config:
                        display_name = self.config[counter].get(
                            'alias', self.config[counter]['column'])
                        data_type = self.STRING_PY_TYPES[
                            self.config[counter]['type']]
                        transform = self.config[counter].get('encoding', None)
                        options = self.config[counter].get('encoding_options', None)
                        expand = self.config[counter].get('expand', False)
                        visible = self.config[counter].get('visible', True)
                        counter += 1
                    else:
                        display_name = row[1]
                        data_type = self.SQLITE_PY_TYPES.get(
                            row[2].upper(), str)
                        transform = None  # TODO: eg. buffer
                        options = None
                        expand = False
                        visible = True

                    col_dict = {
                        'name': col_name,
                        'display': display_name,
                        'type': data_type,
                        'transform': transform,
                        'transform_options': options,
                        'expand': expand,
                        'visible': visible,
                        'from_config': self.config is not None,
                    }

                    if col_name == self.ID_COLUMN:
                        self.id_column_idx = i
                    if col_name == self.PARENT_ID_COLUMN:
                        self.parent_column_idx = i
                    if col_name == self.CHILDREN_LEN_COLUMN:
                        self.children_len_column_idx = i
                    if col_name == self.FLAT_COLUMN:
                        self.flat_column_idx = i
                    if col_name == self.SELECTED_COLUMN:
                        self.selected_column_idx = i
                        has_selected = True
                        col_dict['transform'] = 'boolean'

                    cols.append(col_dict)

                if self._ensure_selected_column and not has_selected:
                    alter_sql = 'ALTER TABLE %s ADD %s INTEGER' % (
                        self.update_table, self.SELECTED_COLUMN)
                    cursor.execute(alter_sql)
                    conn.commit()
                    col_dict = {
                        'name': self.SELECTED_COLUMN,
                        'display': self.SELECTED_COLUMN,
                        'type': int,
                        'transform': 'boolean',
                        'transform_options': 'boolean',
                        'expand': False,
                        'visible': True,
                        'from_config': False,
                    }
                    self.selected_column_idx = len(cols)
                    cols.append(col_dict)

        return cols

    def _load_tree_rows(self, conn, where, order_by, parent_id):
        """Load rows as a tree."""
        if where is not None:
            # FIXME: If we have a where clause, we cant load the results lazily
            # because, we don't know if a row's children/grandchildren/etc will
            # match.  If this optimization (loading the leafs and the necessary
            # parents until the root) good enough?
            children = {}
            node_mapper = {}

            def load_rows(where_):
                query = self.select(
                    conn, self.table, columns=self.table.columns,
                    where=where_, order_by=order_by)
                for row in query:
                    row_id = row[self.id_column_idx]
                    if row_id in node_mapper:
                        continue

                    c_list = children.setdefault(
                        row[self.parent_column_idx], [])
                    node = Node(data=row)
                    c_list.append(node)
                    node_mapper[row_id] = node

            load_rows(where)
            if not children:
                return

            # Load parents incrementally until we are left with the root
            while children.keys() != [None]:
                parents_to_load = []
                for parent, c_list in children.items():
                    if parent is None:
                        continue

                    node = node_mapper.get(parent, None)
                    if node is None:
                        parents_to_load.append(parent)
                        continue

                    node.extend(c_list)
                    node.children_len = len(node)
                    del children[parent]

                if parents_to_load:
                    where = self.table.columns[self.ID_COLUMN].in_(
                        parents_to_load)
                    load_rows(where)

            for node in children[None]:
                yield node
        else:
            # If there's no where clause, we can load the results lazily
            where = self.table.columns[self.PARENT_ID_COLUMN] == parent_id

            if self.CHILDREN_LEN_COLUMN is None:
                count_table = alias(self.table, '__count')
                # We could use the comparison between the columns, but that would
                # make sqlalchemy add self.table in the FROM clause, which
                # would produce wrong results.
                count_where = '%s.%s = %s.%s' % (
                    count_table.name, self.PARENT_ID_COLUMN,
                    self.table.name, self.ID_COLUMN)
                count_select = select(
                    [func.count(1)],
                    whereclause=count_where, from_obj=[count_table])

                columns = self.table.columns.values()
                # We have to compile this here or else sqlalchemy would put
                # this inside the FROM part.
                columns.append('(%s)' % (_compile(count_select), ))
                extra_count_col = True
            else:
                columns = self.table.columns
                extra_count_col = False

            query = self.select(
                conn, self.table, columns=columns,
                where=where, order_by=order_by)

            for row in query:
                if extra_count_col:
                    children_len = row.pop(-1)
                else:
                    children_len = row[self.children_len_column_idx]

                yield Node(data=row, children_len=children_len)


def rank(matchinfo):
    """Rank full-text search results.

    :param matchinfo: defined as returning 32-bit unsigned integers in
      machine byte order (http://www.sqlite.org/fts3.html#matchinfo)
      and struct defaults to machine byte order.
    """
    matchinfo = struct.unpack('I' * (len(matchinfo) / 4), matchinfo)
    iterator = iter(matchinfo[2:])
    return sum(x[0] for x in zip(iterator, iterator, iterator) if x[1])


class Database(object):

    """Generic database object.

    This class is subclassed to provide additional functionality specific to
    artifacts and/or documents.

    :param db_filename: Path to the sqlite database file
    :type db_filename: str

    """

    def __init__(self, db_filename):
        """Connect to database and create session object."""
        self.db_filename = db_filename
        self.engine = create_engine(
            'sqlite:///{}'.format(db_filename),
            connect_args={'check_same_thread': False},
        )
        self.connection = None
        self.metadata = MetaData(bind=self.engine)

    def connect(self):
        """Create connection."""
        logger.debug('Connecting to SQLite database: %s', self.db_filename)
        self.connection = self.engine.connect()

    def disconnect(self):
        """Close connection."""
        assert not self.connection.closed
        logger.debug(
            'Disconnecting from SQLite database: %s', self.db_filename)
        self.connection.close()

    def __enter__(self):
        """Connect on entering context."""
        self.connect()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Disconnect on exiting context."""
        self.disconnect()

    def __getitem__(self, table_name):
        """Get table object in database.

        :param table_name: Name of the table
        :type table_name: str
        :return: Table object that can be used in queries
        :rtype: sqlalchemy.schema.Table

        """
        table = self.metadata.tables.get(table_name)
        if table is None:
            table = Table(table_name, self.metadata, autoload=True)
        return table

    def run_quick_check(self):
        """Check database integrity.

        Some files, especially those files created after carving, might not
        contain completely valid data.

        """
        try:
            result = self.connection.execute('PRAGMA quick_check;')
        except DatabaseError:
            return False

        passed = result.fetchone()[0] == 'ok'
        if not passed:
            logger.warning('Integrity check failure: %s', self.db_filename)
        return passed

    def reflect(self, table_name):
        """Get table metadata through reflection.

        sqlalchemy already provides a reflect method, but it will stop at the
        first failure, while this method will try to get as much as possible.

        """
        inspector = inspect(self.engine)
        columns = []
        for column_data in inspector.get_columns(table_name):
            # Rename 'type' to 'type_' to create column object
            column_type = column_data.pop('type', None)
            column_data['type_'] = column_type
            columns.append(Column(**column_data))

        return Table(table_name, self.metadata, *columns)



# -*- coding: utf-8 -*-

__author__ = 'NowSecure, Inc.'
__email__ = 'info@nowsecure.com'
__version__ = '0.1.5'

import logging
import os

from gi.repository import Gtk


icon_theme = Gtk.IconTheme.get_default()
# Register our icons as ultimate fallback. Those should only
# be used when there's no icon theme registered (e.g. xvfb)
icon_theme.append_search_path(
    os.path.join(os.path.dirname(os.path.abspath(__file__)),
                 'data', 'media', 'icons'))

# Set default logging handler to avoid "No handler found" warnings.
logging.getLogger(__name__).addHandler(logging.NullHandler())



# -*- coding: utf-8 -*-

# NOTE: tests directory can go either inside or outside package directory.
# In cases where the production environment is likely to be quite different
# from the development environment, it can be useful to distribute the
# tests to be able to run them in the production environment.
